"""
This module hosts utilities imported across modules. 
"""

from typing import Callable, ClassVar, Union, TypeVar, Generic, get_type_hints
import datetime

__all__ = ["register_union_type", "Symbol", "S", "OrderedSet", "LITERAL_TYPES", "LITERAL_TYPE_SIG"]


# -----------------------------------------------------------
# data structures used across the package
# -----------------------------------------------------------


class SymbolHelper(type):
    """
    Customize attribute access to double up as a constructor.
    So, `Symbol.foo` is the same as `Symbol("foo")`.
    """

    def __getattr__(cls, key: str) -> "Symbol":
        if key.startswith("_"):
            raise AttributeError(
                "fluent syntax isn't supported for strings starting with an underscore"
            )
        return cls(key)


class Symbol(metaclass=SymbolHelper):
    """
    Use Symbol objects to refer to identifiers, function names, etc.
    Returns the same copy for each type of symbol.
    """

    _instances: ClassVar[dict[str, "Symbol"]] = {}
    data: str

    def __new__(cls, data: Union[str, "Symbol"]):
        if isinstance(data, Symbol):
            return data

        assert isinstance(data, str)
        if data not in cls._instances:
            cls._instances[data] = super().__new__(cls)
        return cls._instances[data]

    def __init__(self, data: Union[str, "Symbol"]):
        if isinstance(data, str):  # else the `data` attribute is already set
            self.data = data

    def __str__(self) -> str:
        return self.data

    def __repr__(self) -> str:
        return self.data

    def __hash__(self) -> int:
        return hash(self.data)


# Shorthand for the Symbol class
S = Symbol

_T = TypeVar("_T")


class OrderedSet(Generic[_T]):
    """We need an ordered set at a bunch of places"""

    _map: dict[_T, bool]

    def __init__(self):
        self._map = {}

    def add(self, item: _T):
        self._map[item] = True

    def pop(self):
        return self._map.popitem()[0]

    def __contains__(self, item: _T):
        return item in self._map

    def __iter__(self):
        return iter(self._map)

    def __len__(self):
        return len(self._map)

    def __repr__(self):
        return repr(self._map)


LITERAL_TYPES = (
    int,
    float,
    str,
    bool,
    type(None),
    datetime.date,
    datetime.time,
    datetime.datetime,
    datetime.timedelta,
)
# unpacking list in a type object subscript only possible since python 3.11
LITERAL_TYPE_SIG = Union[
    int,
    float,
    str,
    bool,
    None,
    datetime.date,
    datetime.time,
    datetime.datetime,
    datetime.timedelta,
]

# -----------------------------------------------------------
# utility routines
# -----------------------------------------------------------


def register_union_type(target_fn: Callable) -> Callable:
    """The singledispatch.register decorator doesn't work for functions that have a
    union type as the first argument. This decorator is a workaround for that.

    credits: Python bug tracker [link](https://bugs.python.org/issue46014)
    """

    def decorator(func):
        arg_type = list(get_type_hints(func).values())[0]
        assert arg_type.__origin__ is Union
        for typ in arg_type.__args__:
            func = target_fn.register(typ)(func)
        return func

    return decorator



"""
This module implements support for rendering to the common SQL variants. 
"""

from ..sqlcontext import SQLDialect
from ..clausedefs import VarStyle, LimitStyle
from ..render import dialect_postgres

__all__ = ["dialect_mysql", "dialect_postgres", "dialect_sqlite"]


def dialect_mysql() -> SQLDialect:
    return SQLDialect(
        name="mysql",
        var_style=VarStyle.POSITIONAL,
        var_prefix="?",
        id_quotes=("`", "`"),
        values_row_constructor="ROW",
        values_column_prefix="column_",
        values_column_index=0,
    )


def dialect_sqlite() -> SQLDialect:
    return SQLDialect(
        name="sqlite",
        var_style=VarStyle.NUMBERED,
        var_prefix="?",
        has_as_columns=False,
        has_datetime_types=False,
    )



import io
from enum import Enum
from functools import lru_cache, singledispatch, wraps
from typing import Any, Iterator, Optional, Type, Union, get_type_hints

from .common import OrderedSet, S, Symbol, LITERAL_TYPES, LITERAL_TYPE_SIG
from .prettier import (
    Doc,
    QuoteContext,
    assg_expr,
    get_screen_width,
    full_query_expr,
    highlight_expr,
    resolve,
    to_doc,
)
from .sqlcontext import SQLTable

__all__ = [
    "SQLNode",
    "NODE_MATERIAL",
    "TabularNode",
    "populate_label_map",
    "label",
    "check_repr_context",
    "ErrDuplicateLabel",
    "ErrIllFormed",
    "ErrRefUndefinedHandle",
    "ErrRefAmbiguousHandle",
    "ErrRefUndefinedName",
    "ErrRefAmbiguousName",
    "ErrRefUnexpectedRowType",
    "ErrRefUnexpectedScalarType",
    "ErrRefUnexpectedAgg",
    "ErrRefAmbiguousAgg",
    "ErrRefUndefinedTable",
    "ErrRefInvalidTable",
]


# -----------------------------------------------------------
# methods exported by this module
# -----------------------------------------------------------


@singledispatch
def label(node: Union["SQLNode", None]) -> Symbol:
    """Returns a default alias for the node; to use when rendering to SQL.
    The implementation for subclasses of SQLNodes is defined in `nodedefs.py`.
    """
    if node is None:
        return S("_")
    else:
        raise NotImplementedError(
            f"label isn't implemented for the SQLNode type - {type(node)}"
        )


def populate_label_map(node: "SQLNode", args: list["SQLNode"]) -> dict[Symbol, int]:
    """Validate that labels of args are unique, for Select and other nodes
    that declare new column references, and cache them.
    """
    label_map = dict[Symbol, int]()
    for i, arg in enumerate(args):
        name = label(arg)
        if name in label_map:
            raise ErrDuplicateLabel(name=name, path=[arg, node])
        label_map[name] = i
    return label_map


# -----------------------------------------------------------
# Definition of SQLNode objects
# -----------------------------------------------------------


class SQLNode:
    """A tabular or scalar operation that can be expressed as a SQL query"""

    highlight: bool  # highlight node when printing the query tree?

    def __init__(self) -> None:
        self.highlight = False

    def __repr__(self) -> str:
        """From the top level node, the tree is traversed and `linearized` to
        produce a representation with exactly one node defined on each line.
        The actual string representation for each node is implemented in the
        `pretty_repr` method.
        """

        # got too long; delegated to another function so the class definition is compact
        return repr_node(self, limit=False)

    def pretty_repr(self, ctx: QuoteContext, full: bool = False) -> "Doc":
        """Implemented by each SQLNode subclass. Returns a Doc object representing
        the SQLNode.
        """
        raise NotImplementedError(
            f"pretty_repr method isn't implemented on the SQLNode class - {type(self)}"
        )

    def rebase(self, pre: "SQLNode") -> "SQLNode":
        raise NotImplementedError(
            f"rebase isn't implemented for the SQLNode class - {type(self)}"
        )

    def __rshift__(self, other: "SQLNode") -> "SQLNode":
        """Used to compose SQL nodes"""
        if isinstance(other, SQLNode):
            return other.rebase(self)
        else:
            raise NotImplementedError(
                f">> isn't a valid operation with a SQLNode and {type(other)}"
            )


class TabularNode(SQLNode):
    """A node that can be used as a SQL subquery"""

    def __init__(self) -> None:
        super().__init__()


# For convenience, user facing SQLNode objects take python literal types as inputs
# and cast them internally to SQLNodes. This represents the allowed types.
NODE_MATERIAL = Union[SQLNode, LITERAL_TYPE_SIG]

# -----------------------------------------------------------
# pretty printing for SQLNode objects
# -----------------------------------------------------------


def _get_origin(typ: Any) -> Any:
    # HACK: looks very fragile, but `get_origin`, `get_args` are not present in python <3.8
    return typ.__origin__ if hasattr(typ, "__origin__") else None


def _get_args(typ: Any) -> tuple:
    return typ.__args__ if hasattr(typ, "__args__") else tuple()


def typ_as_str(data: Any) -> str:
    return data.__class__.__name__


@lru_cache
def get_node_refs(node_typ: Type[SQLNode]) -> tuple[tuple[str], tuple[str]]:
    """Compute attributes on a SQLNode class that point to other SQLNodes. We use
    it to implement visitor pattern for traversing the node tree. Alternatively, we
    could ask each SQLNode to specify them, but that'd be tedious.
    """

    assert issubclass(node_typ, SQLNode)

    maybe_nodes, listof_nodes = [], []
    for attr, typ in get_type_hints(node_typ).items():
        if typ == SQLNode:
            maybe_nodes.append(attr)

        container = _get_origin(typ)
        if container == Union:
            args = _get_args(typ)
            if len(args) == 2 and SQLNode in args and type(None) in args:
                maybe_nodes.append(attr)
        elif container == list and SQLNode in _get_args(typ):
            listof_nodes.append(attr)

    # we want the parent node to be visited first, which the `over` attribute points to
    if "over" in maybe_nodes:
        maybe_nodes.remove("over")
        maybe_nodes.insert(0, "over")
    return (tuple(maybe_nodes), tuple(listof_nodes))


def visit(node: SQLNode, visiting: set[SQLNode]) -> Iterator[SQLNode]:
    """recursively visit attributes that are SQLNodes."""
    if node in visiting:
        return
    visiting.add(node)

    maybe_nodes, listof_nodes = get_node_refs(type(node))
    for attr in maybe_nodes:
        _value = getattr(node, attr)
        if _value is not None:
            yield from visit(_value, visiting)
    for attr in listof_nodes:
        for _value in getattr(node, attr):
            yield from visit(_value, visiting)

    yield node
    visiting.remove(node)


def repr_node(node: SQLNode, limit: bool = False) -> str:
    """Returns a string representation of the SQLNode, with the full tree of
    nodes, linearized. That is, each table reference, tabular node, or a
    repeated node is put on a separate line. While, any references to them are
    replaced by an alias.
    """
    if limit:
        ctx = QuoteContext(limit=True)
        return resolve(to_doc(node, ctx), get_screen_width())

    tables_seen = OrderedSet[SQLTable]()
    nodes_seen = OrderedSet[SQLNode]()
    nodes_toplevel = set[SQLNode]()

    visited: set[SQLNode] = set()
    for ref in visit(node, visited):
        # HACK: we really want to do an isinstance check, but that introduces a cyclic dep.
        if typ_as_str(ref) == "From":
            if isinstance(ref.source, SQLTable):  # type: ignore
                tables_seen.add(ref.source)  # type: ignore
        elif typ_as_str(ref) == "FromTable":
            tables_seen.add(ref.table)  # type: ignore

        if isinstance(ref, TabularNode) or ref in nodes_seen:
            nodes_toplevel.add(ref)
        nodes_seen.add(ref)

    ctx = QuoteContext()
    defs = []
    if len(nodes_toplevel) >= 2 or (
        len(nodes_toplevel) == 1 and node not in nodes_toplevel
    ):
        # start by defining all the table references
        _ctx_limit = QuoteContext(limit=True)
        for tab in tables_seen:
            name = tab.name
            defs.append(assg_expr(str(name), to_doc(tab, _ctx_limit)))
            ctx.vars_[tab] = name

        # add aliases for top level nodes
        q_idx = 0
        for n in nodes_seen:
            if n in nodes_toplevel:
                q_idx += 1
                name = S(f"q{q_idx}")
                ctx.vars_[n] = name

        # add definitions for all top level nodes; child references will be replaced with aliases
        q_idx = 0
        for n in nodes_seen:
            if n in nodes_toplevel:
                q_idx += 1
                name = f"q{q_idx}"
                defs.append(assg_expr(str(name), n.pretty_repr(ctx, full=True)))

    ex = to_doc(node, ctx)
    if len(defs) > 0:
        ex = full_query_expr([*defs, ex])
    return resolve(ex, 80)


def check_repr_context(repr_func):
    """Decorator for the `pretty_repr` method on SQLNode objects. Looks into the
    Quote context to decide if the full expression needs to be printed or an alias
    with do.
    """

    @wraps(repr_func)
    def wrapped(node: SQLNode, ctx: QuoteContext, full: bool = False) -> Doc:
        if ctx.limit:
            return "..."
        else:
            var = ctx.vars_.get(node, None)
            if var is None or full:
                expr = repr_func(node, ctx)
                if node.highlight:
                    return highlight_expr(expr)
                else:
                    return expr
            else:
                return str(var)

    return wrapped


# -----------------------------------------------------------
# Errors to raise when parsing the SQL node tree
# -----------------------------------------------------------


class ErrFunSQL(BaseException):
    name: Optional[Symbol]
    path: list[SQLNode]

    def __init__(
        self,
        name: Union[str, Symbol, None] = None,
        path: Optional[list[SQLNode]] = None,
    ) -> None:
        self.name = S(name) if name is not None else None
        self.path = path if path is not None else []

    def _custom_msg(self) -> str:
        raise NotImplementedError("Custom message should be implemented in subclasses")

    def __str__(self) -> str:
        buf = io.StringIO()
        buf.write(self._custom_msg() + "\n")

        if len(self.path) > 0:
            self.path[0].highlight = True  # highlight the error site
            buf.write(f"\n{self.path[-1]}\n")
        elif isinstance(self, ErrIllFormed):
            buf.write("Top level node\n")
        else:
            buf.write("\n")
        return buf.getvalue()


class ErrDuplicateLabel(ErrFunSQL):
    def _custom_msg(self) -> str:
        return f"FunSQL.ErrDuplicateLabel: {self.name} is used more than once in:"


class ErrIllFormed(ErrFunSQL):
    def _custom_msg(self) -> str:
        return "FunSQL.ErrIllFormed: ill-formed query in:"


class ErrRefUndefinedHandle(ErrFunSQL):
    def _custom_msg(self) -> str:
        return f"FunSQL.ErrReference: node bound reference failed to resolve in:"


class ErrRefAmbiguousHandle(ErrFunSQL):
    def _custom_msg(self) -> str:
        return f"FunSQL.ErrReference: node bound reference is ambiguous in:"


class ErrRefUndefinedName(ErrFunSQL):
    def _custom_msg(self) -> str:
        return f"FunSQL.ErrReference: name {self.name} is undefined in:"


class ErrRefAmbiguousName(ErrFunSQL):
    def _custom_msg(self) -> str:
        return f"FunSQL.ErrReference: name {self.name} is ambiguous in:"


class ErrRefUnexpectedRowType(ErrFunSQL):
    def _custom_msg(self) -> str:
        return f"FunSQL.ErrReference: incomplete reference {self.name} in:"


class ErrRefUnexpectedScalarType(ErrFunSQL):
    def _custom_msg(self) -> str:
        return f"FunSQL.ErrReference: unexpected reference after {self.name} in:"


class ErrRefUnexpectedAgg(ErrFunSQL):
    def _custom_msg(self) -> str:
        return f"FunSQL.ErrReference: aggregate expression allowed only inside a Group or Partition in:"


class ErrRefAmbiguousAgg(ErrFunSQL):
    def _custom_msg(self) -> str:
        return f"FunSQL.ErrReference: ambiguous aggregate expression in:"


class ErrRefUndefinedTable(ErrFunSQL):
    def _custom_msg(self) -> str:
        return f"FunSQL.ErrReference: table reference {self.name} is undefined in:"


class ErrRefInvalidTable(ErrFunSQL):
    def _custom_msg(self) -> str:
        return f"FunSQL.ErrReference: table reference {self.name} required label using `As` in:"



"""
Sections in this module
* utility functions used within this module
* clause definitions
* utility functions for external use
"""

from enum import Enum
from typing import Any, Optional, Union, Literal

from .common import Symbol, S
from .sqlcontext import LimitStyle, VarStyle
from .compiler.serialize import SerializationContext, serialize
from .clauses import SQLClause
from .prettier import (
    Doc,
    QuoteContext,
    call_expr,
    assg_expr,
    list_expr,
    pipe_expr,
    to_doc,
)

__all__ = [
    "AGG",
    "AS",
    "CASE",
    "FROM",
    "FUN",
    "GROUP",
    "HAVING",
    "ID",
    "JOIN",
    "KW",
    "LIMIT",
    "LIT",
    "NOTE",
    "OP",
    "ORDER",
    "Frame",
    "FrameMode",
    "FrameExclude",
    "FrameEdge",
    "FrameEdgeSide",
    "PARTITION",
    "SelectTop",
    "SELECT",
    "ValueOrder",
    "NullsOrder",
    "SORT",
    "UNION",
    "VALUES",
    "VAR",
    "WHERE",
    "WINDOW",
    "WITH",
    "qual",
    "alias",
]

# -----------------------------------------------------------
# Serialization methods for SQLclause objects
# -----------------------------------------------------------


@serialize.register(list)
def _(data: list[Any], ctx: SerializationContext, sep: Optional[str] = None) -> None:
    """Serialize a list of SQL clauses.

    NOTE: Doesn't leave a space before adding characters, do it in the caller.
    """
    sep = ", " if sep is None else f" {sep} "
    if len(data) > 0:
        for i, arg in enumerate(data):
            assert isinstance(arg, (Symbol, SQLClause))
            if i > 0:
                if isinstance(arg, KW):
                    ctx.write(" ")
                else:
                    ctx.write(sep)
            serialize(arg, ctx)


def serialize_lines(
    data: list[SQLClause], ctx: SerializationContext, sep: Optional[str] = None
) -> None:
    """Serialize a list of SQL clauses, one per line. For example, when listing
    multiple column names with a SELECT or ORDER clause.
    """
    if len(data) == 0:
        return
    if len(data) == 1:
        ctx.write(" ")
        serialize(data[0], ctx)
    else:
        ctx.level += 1
        for i, clause in enumerate(data):
            if i > 0:
                if sep is None:
                    ctx.write(", ")
                else:
                    ctx.write(f" {sep} ")
            ctx.newline()
            serialize(clause, ctx)
        ctx.level -= 1


# -----------------------------------------------------------
# utility functions local to this module
# -----------------------------------------------------------

# Type of inputs that need to be converted to a SQLClause
TYPE_INPUT_CLAUSE = Union[SQLClause, Symbol, Any]


def _cast_to_clause(arg: Union[SQLClause, Symbol, Any]) -> SQLClause:
    """Convert a value to a SQLClause."""
    if isinstance(arg, SQLClause):
        return arg
    elif isinstance(arg, Symbol):
        return ID(arg) if str(arg) != "*" else OP(arg)
    else:
        return LIT(arg)


def _cast_to_clause_skip_none(
    arg: Union[SQLClause, Symbol, Any]
) -> Union[SQLClause, None]:
    """Convert a value to a SQLClause, skipping None which otherwise gets
    treated as the NULL literal.
    """
    return None if arg is None else _cast_to_clause(arg)


def _rebase_clause(curr: Optional[SQLClause], pre: SQLClause) -> SQLClause:
    """Shorthand for a common pattern"""
    if curr is None:
        return pre
    else:
        return curr.rebase(pre)


# -----------------------------------------------------------
# SQL clause definitions
# -----------------------------------------------------------


class AGG(SQLClause):
    name: Symbol
    distinct: bool
    args: list[SQLClause]
    filter_: Optional[SQLClause]
    over: Optional[SQLClause]

    def __init__(
        self,
        name: Union[Symbol, str],
        *args: SQLClause,
        distinct: bool = False,
        filter_: Optional[SQLClause] = None,
        over: Optional[SQLClause] = None,
    ) -> None:
        self.name = S(name)
        self.args = [_cast_to_clause(arg) for arg in args]
        self.distinct = distinct
        self.filter_ = filter_
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "AGG"
        args = []

        args.append(to_doc(str(self.name)))
        if self.distinct:
            args.append(assg_expr("distinct", to_doc(self.distinct)))
        for arg in self.args:
            args.append(to_doc(arg))
        if self.filter_ is not None:
            args.append(assg_expr("filter", to_doc(self.filter_)))
        if self.over is not None:
            args.append(assg_expr("over", to_doc(self.over)))

        return call_expr(name, args)

    def rebase(self, pre: SQLClause) -> "AGG":
        return self.__class__(
            self.name,
            *self.args,
            distinct=self.distinct,
            filter_=self.filter_,
            over=_rebase_clause(self.over, pre),
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        has_filter = self.filter_ is not None
        has_over = self.over is not None

        if has_filter or has_over:
            ctx.write("(")

        ctx.write(str(self.name))
        with ctx.parens():
            if self.distinct:
                ctx.write("DISTINCT ")
            serialize(self.args, ctx)
        if has_filter:
            ctx.write(" FILTER ")
            with ctx.parens():
                ctx.write("WHERE ")
                serialize(self.filter_, ctx)
        if has_over:
            ctx.write(" OVER ")
            with ctx.parens():
                serialize(self.over, ctx)

        if has_filter or has_over:
            ctx.write(")")


class AS(SQLClause):
    name: Symbol
    columns: Optional[list[Symbol]]
    over: Optional[SQLClause]

    def __init__(
        self,
        name: Symbol,
        columns: Optional[list[Symbol]] = None,
        over: Optional[SQLClause] = None,
    ) -> None:
        self.name = name
        self.columns = None if columns is None else [S(c) for c in columns]
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "AS"
        args: list["Doc"] = [str(self.name)]
        if self.columns is not None:
            args.append(
                assg_expr("columns", list_expr([to_doc(col) for col in self.columns]))
            )

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "AS":
        return self.__class__(
            name=self.name, columns=self.columns, over=_rebase_clause(self.over, pre)
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        if isinstance(self.over, PARTITION):
            assert (
                self.columns is None
            ), "PARTITION clause can't be aliased as a table with columns"
            serialize(self.name, ctx)
            ctx.write(" AS ")
            with ctx.parens():
                serialize(self.over, ctx)
        elif self.over is not None:
            serialize(self.over, ctx)
            ctx.write(" AS ")
            serialize(self.name, ctx)
            if self.columns is not None:
                with ctx.parens(space=True):
                    serialize(self.columns, ctx)


class CASE(SQLClause):
    args: list[SQLClause]

    def __init__(self, *args: SQLClause) -> None:
        self.args = [_cast_to_clause(arg) for arg in args]

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "CASE"
        args: list["Doc"] = []

        if len(self.args) < 2:
            args.append(
                assg_expr("args", list_expr([to_doc(arg) for arg in self.args]))
            )
        else:
            args.extend([to_doc(arg) for arg in self.args])
        return call_expr(name, args)

    def _serialize(self, ctx: SerializationContext) -> None:
        with ctx.parens():
            ctx.write("CASE")
            nargs = len(self.args)
            for i, arg in enumerate(self.args):
                if i % 2 == 0:
                    ctx.write(" WHEN " if i < nargs - 1 else " ELSE ")
                else:
                    ctx.write(" THEN ")
                serialize(arg, ctx)
            ctx.write(" END")


class FROM(SQLClause):
    over: Optional[SQLClause] = None

    def __init__(self, over: Optional[Union[SQLClause, Symbol]] = None) -> None:
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "FROM"
        ex = call_expr(name, [])

        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "FROM":
        return self.__class__(over=_rebase_clause(self.over, pre))

    def _serialize(self, ctx: SerializationContext) -> None:
        ctx.newline()
        ctx.write("FROM")
        if self.over is not None:
            ctx.write(" ")
            serialize(self.over, ctx)


class FUN(SQLClause):
    """A SQL Function"""

    name: Symbol
    args: list[SQLClause]

    def __init__(self, name: Union[Symbol, str], *args: SQLClause) -> None:
        self.name = S(name)
        self.args = [_cast_to_clause(arg) for arg in args]

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "FUN"
        args = [to_doc(str(self.name))]
        for arg in self.args:
            args.append(to_doc(arg))

        ex = call_expr(name, args)
        return ex

    def _serialize(self, ctx: SerializationContext) -> None:
        ctx.write(str(self.name))
        with ctx.parens():
            serialize(self.args, ctx)


class GROUP(SQLClause):
    by: list[SQLClause]
    over: Optional[SQLClause]

    def __init__(self, *by: SQLClause, over: Optional[SQLClause] = None) -> None:
        self.by = [_cast_to_clause(arg) for arg in by]
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        ex = call_expr("GROUP", [to_doc(arg) for arg in self.by])
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "GROUP":
        return self.__class__(*self.by, over=_rebase_clause(self.over, pre))

    def _serialize(self, ctx: SerializationContext) -> None:
        if self.over is not None:
            serialize(self.over, ctx)
        if len(self.by) > 0:
            ctx.newline()
            ctx.write("GROUP BY")
            serialize_lines(self.by, ctx)


class HAVING(SQLClause):
    condition: SQLClause
    over: Optional[SQLClause]

    def __init__(self, condition: SQLClause, over: Optional[SQLClause] = None) -> None:
        self.condition = condition
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        ex = call_expr("HAVING", [to_doc(self.condition)])
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "HAVING":
        return self.__class__(
            condition=self.condition, over=_rebase_clause(self.over, pre)
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        if self.over is not None:
            serialize(self.over, ctx)
        ctx.newline()
        ctx.write("HAVING")

        cond = self.condition
        if isinstance(cond, OP) and cond.name == "AND" and len(cond.args) >= 2:
            serialize_lines(cond.args, ctx, sep="AND")
        else:
            ctx.write(" ")
            serialize(cond, ctx)


class ID(SQLClause):
    name: Symbol
    over: Optional[SQLClause] = None

    def __init__(self, name: Symbol, over: Optional[SQLClause] = None) -> None:
        assert isinstance(
            name, Symbol
        ), f"Invalid type to initialize a Symbol: {type(name)}"
        self.name = name
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        ex = call_expr("ID", [to_doc(self.name)])
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "ID":
        return self.__class__(name=self.name, over=_rebase_clause(self.over, pre))

    def _serialize(self, ctx: SerializationContext) -> None:
        if self.over is not None:
            serialize(self.over, ctx)
            ctx.write(".")
        serialize(self.name, ctx)


class JOIN(SQLClause):
    """Your regular SQL Join clause"""

    joinee: SQLClause
    on: SQLClause
    left: bool = False
    right: bool = False
    lateral: bool = False
    over: Optional[SQLClause] = None

    def __init__(
        self,
        joinee: SQLClause,
        on: SQLClause,
        left: bool = False,
        right: bool = False,
        lateral: bool = False,
        over: Optional[SQLClause] = None,
    ) -> None:
        self.joinee = _cast_to_clause(joinee)
        self.on = _cast_to_clause(on)
        self.left = left
        self.right = right
        self.lateral = lateral
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "JOIN"
        args = []

        args.append(to_doc(self.joinee))
        args.append(to_doc(self.on))
        if self.left:
            args.append(assg_expr("left", to_doc(True)))
        if self.right:
            args.append(assg_expr("right", to_doc(True)))
        if self.lateral:
            args.append(assg_expr("lateral", to_doc(True)))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "JOIN":
        return self.__class__(
            joinee=self.joinee,
            on=self.on,
            left=self.left,
            right=self.right,
            lateral=self.lateral,
            over=_rebase_clause(self.over, pre),
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        if self.over is not None:
            serialize(self.over, ctx)
        ctx.newline()

        is_cross = (
            (not self.left)
            and (not self.right)
            and (isinstance(self.on, LIT) and self.on.val == True)
        )
        if is_cross:
            ctx.write("CROSS JOIN ")
        elif self.left and self.right:
            ctx.write("FULL JOIN ")
        elif self.left:
            ctx.write("LEFT JOIN ")
        elif self.right:
            ctx.write("RIGHT JOIN ")
        else:
            ctx.write("INNER JOIN ")  # default JOIN type

        if self.lateral:
            ctx.write("LATERAL ")

        serialize(self.joinee, ctx)
        if not is_cross:
            ctx.write(" ON ")
            serialize(self.on, ctx)


class KW(SQLClause):
    """Keyword argument of a function/operator"""

    name: Symbol
    over: Optional[SQLClause] = None

    def __init__(self, name: Symbol, over: Optional[SQLClause] = None) -> None:
        self.name = name
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        ex = call_expr("KW", [str(self.name)])
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "KW":
        return self.__class__(name=self.name, over=_rebase_clause(self.over, pre))

    def _serialize(self, ctx: SerializationContext) -> None:
        ctx.write(str(self.name))
        if self.over is not None:
            ctx.write(" ")
            serialize(self.over, ctx)


class LIMIT(SQLClause):
    """
    NOTE: `with_ties` is not a supported construct in SQLite or MYSQL dialects
    """

    over: Optional[SQLClause]
    offset: Optional[int]
    limit: Optional[int]
    with_ties: bool

    def __init__(
        self,
        limit: Optional[int] = None,
        *,
        offset: Optional[int] = None,
        with_ties: bool = False,
        over: Optional[SQLClause] = None,
    ) -> None:
        """Can be initialized as LIMIT(:offset, :limit) or LIMIT(:limit)"""
        self.limit = limit
        self.offset = offset
        self.with_ties = with_ties
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "LIMIT"
        args = []

        if self.limit is not None:
            args.append(to_doc(self.limit))
        if self.offset is not None:
            args.append(assg_expr("offset", to_doc(self.offset)))
        if self.with_ties:
            args.append(assg_expr("with_ties", to_doc(True)))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "LIMIT":
        return self.__class__(
            self.limit,
            offset=self.offset,
            over=_rebase_clause(self.over, pre),
            with_ties=self.with_ties,
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        if self.over is not None:
            serialize(self.over, ctx)
        if self.offset is None and self.limit is None:
            return

        if ctx.dialect.limit_style == LimitStyle.REGULAR:
            ctx.newline()
            ctx.write("LIMIT ")
            ctx.write(f"{self.limit}" if self.limit is not None else "-1")
            if self.offset is not None:
                ctx.newline()
                ctx.write("OFFSET ")
                ctx.write(f"{self.offset}")
        else:
            # the FETCH FIRST N ROWS syntax
            if self.offset is not None:
                ctx.newline()
                ctx.write("OFFSET ")
                ctx.write(f"{self.offset}")
                ctx.write("ROW" if self.offset == 1 else "ROWS")
            if self.limit is not None:
                ctx.newline()
                ctx.write("FETCH FIRST " if self.offset is None else "FETCH NEXT ")
                ctx.write(f"{self.limit}")
                ctx.write(" ROW" if self.limit == 1 else " ROWS")
                ctx.write(" WITH TIES" if self.with_ties else " ONLY")


class LIT(SQLClause):
    val: Any

    def __init__(self, val: Any) -> None:
        self.val = val

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        val = to_doc(self.val)
        return f"LIT({val})"

    def _serialize(self, ctx: SerializationContext) -> None:
        serialize(self.val, ctx)


class NOTE(SQLClause):
    text: str
    postfix: bool
    over: Optional[SQLClause]

    def __init__(
        self, text: str, postfix: bool = False, over: Optional[SQLClause] = None
    ) -> None:
        self.text = text
        self.postfix = postfix
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "NOTE"
        args = []

        args.append(to_doc(self.text))
        if self.postfix:
            args.append(assg_expr("postfix", to_doc(True)))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "NOTE":
        return self.__class__(
            text=self.text, over=_rebase_clause(self.over, pre), postfix=self.postfix
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        if self.over is None:
            ctx.write(self.text)
        elif self.postfix:
            serialize(self.over, ctx)
            ctx.write(f" {self.text}")
        else:
            ctx.write(f"{self.text} ")
            serialize(self.over, ctx)


class OP(SQLClause):
    """An SQL operator"""

    name: Symbol
    args: list[SQLClause]

    def __init__(self, name: Union[Symbol, str], *args: TYPE_INPUT_CLAUSE) -> None:
        self.name = S(name)
        self.args = [_cast_to_clause(x) for x in args]

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "OP"
        args = [to_doc(str(self.name))]
        for arg in self.args:
            args.append(to_doc(arg))

        ex = call_expr(name, args)
        return ex

    def _serialize(self, ctx: SerializationContext) -> None:
        name_str = str(self.name)
        if len(self.args) == 0:
            ctx.write(name_str)
        elif len(self.args) == 1:
            with ctx.parens():
                ctx.write(f"{name_str} ")
                serialize(self.args[0], ctx)
        else:
            with ctx.parens():
                serialize(self.args, ctx, sep=name_str)


class ORDER(SQLClause):
    """SQL ORDER BY clause"""

    by: list[SQLClause]
    over: Optional[SQLClause]

    def __init__(self, *by: SQLClause, over: Optional[SQLClause] = None) -> None:
        self.by = [_cast_to_clause(x) for x in by]
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "ORDER"
        args = []
        for arg in self.by:
            args.append(to_doc(arg))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "ORDER":
        return self.__class__(*self.by, over=_rebase_clause(self.over, pre))

    def _serialize(self, ctx: SerializationContext) -> None:
        if self.over is not None:
            serialize(self.over, ctx)
        if len(self.by) > 0:
            ctx.newline()
            ctx.write("ORDER BY")
            serialize_lines(self.by, ctx)


class FrameMode(Enum):
    """Frame mode for window functions"""

    RANGE = "RANGE"
    ROWS = "ROWS"
    GROUPS = "GROUPS"


class FrameExclude(Enum):
    """Frame exclusion for window functions per row/group"""

    NO_OTHERS = "NO OTHERS"
    CURRENT_ROW = "CURRENT ROW"
    GROUP = "GROUP"
    TIES = "TIES"


class FrameEdgeSide(Enum):
    PRECEDING = "PRECEDING"
    CURRENT_ROW = "CURRENT ROW"
    FOLLOWING = "FOLLOWING"


class FrameEdge:
    """
    Corresponds to the `frame_start` or `frame_end` boundary in the partition clause.
    Specified as: `{val} {typ}`

    Args:
        typ: {FrameEdgeSide}
        val: None - unbounded, any other expr - {expr}
    """

    typ: FrameEdgeSide
    val: Union[None, Any]

    def __init__(self, typ: FrameEdgeSide, val: Union[None, Any] = None) -> None:
        self.typ = typ
        self.val = val

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        if self.typ == FrameEdgeSide.CURRENT_ROW:
            return "CURRENT_ROW"
        else:
            return call_expr(self.typ.value, [to_doc(self.val)])


class Frame:
    """Specify the partition fram for a window function. Expanded by the SQL query,
    it corresponds to:

    `{ mode } BETWEEN { start } AND { end } [ exclude ]`

    Args:
        mode: The frame mode
        start: The start boundary
        end: The end boundary

    NOTE: The rules for default values of (frame_start, frame_end) depend on if the
    order_by clause in the corresponding partition.
    - (None, 0) if order_by is present
    - (None, None) if not

    This confuses me, so we make both mandatory if you're using a partition frame.
    """

    mode: FrameMode
    start: FrameEdge
    end: FrameEdge
    exclude: Optional[FrameExclude] = None

    def __init__(
        self,
        mode: FrameMode,
        start: FrameEdge,
        end: FrameEdge,
        exclude: Optional[FrameExclude] = None,
    ) -> None:
        self.mode = mode
        self.start = start
        self.end = end
        self.exclude = exclude

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        if self.start is None and self.end is None and self.exclude is None:
            return self.mode.value

        list_ex = []
        list_ex.append(assg_expr("mode", self.mode.value))
        if self.start is not None:
            list_ex.append(assg_expr("start", to_doc(self.start)))
        if self.end is not None:
            list_ex.append(assg_expr("end", to_doc(self.end)))
        if self.exclude is not None:
            list_ex.append(assg_expr("exclude", self.exclude.value))

        return list_expr(list_ex)

    def _serialize_boundary(
        self, ctx: SerializationContext, boundary: FrameEdge
    ) -> None:
        if boundary.typ == FrameEdgeSide.CURRENT_ROW:
            ctx.write("CURRENT ROW")
        elif boundary.val is None:
            ctx.write(f"UNBOUNDED {boundary.typ.value}")
        else:
            serialize(boundary.val, ctx)
            ctx.write(f" {boundary.typ.value}")

    def _serialize(self, ctx: SerializationContext) -> None:
        ctx.write(f"{self.mode.value} ")

        ctx.write("BETWEEN ")
        self._serialize_boundary(ctx, self.start)
        ctx.write(" AND ")
        self._serialize_boundary(ctx, self.end)
        if self.exclude is not None:
            ctx.write(f" EXCLUDE {self.exclude.value}")


class PARTITION(SQLClause):
    """
    Look at the postgres and sqlite documentation to figure out the semantics
    - sqlite: https://www.sqlite.org/windowfunctions.html
    - postgres: https://www.postgresql.org/docs/14/sql-expressions.html#SYNTAX-WINDOW-FUNCTIONS
    """

    by: list[SQLClause]
    order_by: list[SQLClause]
    frame: Optional[Frame]
    over: Optional[SQLClause]

    def __init__(
        self,
        *by: SQLClause,
        order_by: Optional[list[SQLClause]] = None,
        frame: Optional[Frame] = None,
        over: Optional[SQLClause] = None,
    ) -> None:
        self.by = [_cast_to_clause(x) for x in by]
        self.order_by = (
            [] if order_by is None else [_cast_to_clause(x) for x in order_by]
        )
        self.frame = frame
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "PARTITION"
        args = []
        for a in self.by:
            args.append(to_doc(a))
        if len(self.order_by) > 0:
            args.append(
                assg_expr("order_by", list_expr([to_doc(a) for a in self.order_by]))
            )
        if self.frame is not None:
            args.append(assg_expr("frame", to_doc(self.frame)))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: "SQLClause") -> "PARTITION":
        return self.__class__(
            *self.by,
            order_by=self.order_by,
            frame=self.frame,
            over=_rebase_clause(self.over, pre),
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        add_space: bool = False
        if self.over is not None:
            add_space = True
            serialize(self.over, ctx)
        if len(self.by) > 0:
            ctx.write(" " if add_space else "")
            add_space = True
            ctx.write("PARTITION BY ")
            serialize(self.by, ctx)
        if len(self.order_by) > 0:
            ctx.write(" " if add_space else "")
            add_space = True
            ctx.write("ORDER BY ")
            serialize(self.order_by, ctx)
        if self.frame is not None:
            ctx.write(" " if add_space else "")
            serialize(self.frame, ctx)


class SelectTop(SQLClause):
    """When selecting top n rows"""

    limit: int
    with_ties: bool

    def __init__(self, limit: int, with_ties: bool = False) -> None:
        self.limit = limit
        self.with_ties = with_ties

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        if not self.with_ties:
            return to_doc(self.limit)
        else:
            return list_expr(
                [
                    assg_expr("limit", to_doc(self.limit)),
                    assg_expr("with ties", to_doc(self.with_ties)),
                ]
            )


class SELECT(SQLClause):
    """SQL SELECT clause where it all comes together"""

    args: list[SQLClause]
    distinct: bool
    top: Optional[SelectTop]
    over: Optional[SQLClause]

    def __init__(
        self,
        *args: Union[SQLClause, Symbol, Any],
        distinct: bool = False,
        top: Optional[SelectTop] = None,
        over: Optional[SQLClause] = None,
    ) -> None:
        self.args = [_cast_to_clause(x) for x in args]
        self.distinct = distinct
        self.top = top
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "SELECT"
        args = []

        if self.distinct:
            args.append(assg_expr("distinct", to_doc(self.distinct)))
        if self.top is not None:
            args.append(assg_expr("top", to_doc(self.top)))
        if len(self.args) == 0:
            args.append(assg_expr("args", to_doc("[]")))
        else:
            for arg in self.args:
                args.append(to_doc(arg))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "SELECT":
        return self.__class__(
            *self.args,
            distinct=self.distinct,
            top=self.top,
            over=_rebase_clause(self.over, pre),
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        nested_orig = ctx.nested
        if nested_orig:
            ctx.level += 1
            ctx.write("(")
            ctx.newline()

        ctx.nested = True
        ctx.write("SELECT")
        if self.top is not None:
            ctx.write(f" TOP {self.top.limit}")
            if self.top.with_ties:
                ctx.write(" WITH TIES")

        if self.distinct:
            ctx.write(" DISTINCT")
        serialize_lines(self.args, ctx)
        if self.over is not None:
            serialize(self.over, ctx)

        ctx.nested = nested_orig
        if nested_orig:
            ctx.level -= 1
            ctx.newline()
            ctx.write(")")


class ValueOrder(Enum):
    ASC = "ASC"
    DESC = "DESC"


class NullsOrder(Enum):
    FIRST = "FIRST"
    LAST = "LAST"


class SORT(SQLClause):
    value: ValueOrder
    nulls: Optional[NullsOrder] = None
    over: Optional[SQLClause] = None

    def __init__(
        self,
        value: ValueOrder,
        nulls: Optional[NullsOrder] = None,
        over: Optional[SQLClause] = None,
    ) -> None:
        self.value = value
        self.nulls = nulls
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "SORT"
        args = []

        args.append(self.value.value)
        if self.nulls is not None:
            args.append(assg_expr("nulls", self.nulls.value))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "SORT":
        return self.__class__(
            value=self.value, nulls=self.nulls, over=_rebase_clause(self.over, pre)
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        if self.over is not None:
            serialize(self.over, ctx)
            ctx.write(" ")
        ctx.write(self.value.value)
        if self.nulls is not None:
            ctx.write(" NULLS " + self.nulls.value)


class UNION(SQLClause):
    args: list[SQLClause]
    all_: bool
    over: Optional[SQLClause]

    def __init__(
        self,
        *args: SQLClause,
        all_: bool = False,
        over: Optional[SQLClause] = None,
    ) -> None:
        self.args = [_cast_to_clause(x) for x in args]
        self.all_ = all_
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "UNION"
        args = []

        if self.all_:
            args.append(assg_expr("all", to_doc(self.all_)))
        if len(self.args) == 0:
            args.append(assg_expr("args", "[]"))
        else:
            for arg in self.args:
                args.append(to_doc(arg))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "UNION":
        return self.__class__(
            *self.args, all_=self.all_, over=_rebase_clause(self.over, pre)
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        nested_orig = ctx.nested
        if nested_orig:
            ctx.level += 1
            ctx.write("(")
            ctx.newline()

        ctx.nested = False
        if self.over is not None:
            serialize(self.over, ctx)
        for arg in self.args:
            ctx.newline()
            ctx.write("UNION" if not self.all_ else "UNION ALL")
            ctx.newline()
            serialize(arg, ctx)

        ctx.nested = nested_orig
        if nested_orig:
            ctx.level -= 1
            ctx.newline()
            ctx.write(")")


class VALUES(SQLClause):
    rows: list[Any]

    def __init__(self, rows: list[Union[Any, tuple]]) -> None:
        self.rows = rows

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "VALUES"
        args = []

        make_tuple = lambda seq: ", ".join([to_doc(x) for x in seq])  # type: ignore
        if isinstance(self.rows[0], tuple):
            args.append(list_expr([f"({make_tuple(row)})" for row in self.rows]))
        else:
            args.append(make_tuple(self.rows))
        ex = call_expr(name, args)
        return ex

    def _serialize(self, ctx: SerializationContext) -> None:
        nested_orig = ctx.nested
        if nested_orig:
            ctx.level += 1
            ctx.write("(")
            ctx.newline()

        ctx.nested = True
        ctx.write("VALUES")
        l = len(self.rows)
        if l == 1:
            ctx.write(" ")
        elif l > 1:
            ctx.level += 1
            ctx.newline()

        row_prefix = ctx.dialect.values_row_constructor
        for i, row in enumerate(self.rows):
            if i > 0:
                ctx.write(",")
                ctx.newline()
            if isinstance(row, tuple):
                if row_prefix is not None:
                    ctx.write(row_prefix)

                with ctx.parens():
                    for j, val in enumerate(row):
                        if j > 0:
                            ctx.write(", ")
                        serialize(val, ctx)
            else:
                serialize(row, ctx)

        if l > 1:
            ctx.level -= 1
        ctx.nested = nested_orig
        if nested_orig:
            ctx.level -= 1
            ctx.newline()
            ctx.write(")")


class VAR(SQLClause):
    name: Symbol

    def __init__(self, name: Symbol) -> None:
        self.name = name

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        return call_expr("VAR", [str(self.name)])

    def _serialize(self, ctx: SerializationContext) -> None:
        ctx.write(ctx.dialect.var_prefix)
        style = ctx.dialect.var_style

        if style == VarStyle.POSITIONAL:
            ctx.variables.append(self.name)
        else:
            try:
                pos = ctx.variables.index(self.name)
            except ValueError:
                pos = len(ctx.variables)
                ctx.variables.append(self.name)
            if style == VarStyle.NAMED:
                ctx.write(str(self.name))
            elif style == VarStyle.NUMBERED:
                ctx.write(str(1 + pos))


class WHERE(SQLClause):
    condition: SQLClause
    over: Optional[SQLClause] = None

    def __init__(self, condition: SQLClause, over: Optional[SQLClause] = None) -> None:
        self.condition = _cast_to_clause(condition)
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "WHERE"
        args = [to_doc(self.condition)]

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "WHERE":
        return self.__class__(
            condition=self.condition,
            over=_rebase_clause(self.over, pre),
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        if self.over is not None:
            serialize(self.over, ctx)

        ctx.newline()
        ctx.write("WHERE")
        cond = self.condition
        if isinstance(cond, OP) and cond.name == "AND" and len(cond.args) >= 2:
            serialize_lines(cond.args, ctx, sep="AND")
        else:
            ctx.write(" ")
            serialize(cond, ctx)


class WINDOW(SQLClause):
    args: list[SQLClause]
    over: Optional[SQLClause]

    def __init__(self, *args: SQLClause, over: Optional[SQLClause] = None) -> None:
        self.args = [_cast_to_clause(arg) for arg in args]
        self.over = _cast_to_clause_skip_none(over)

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "WINDOW"
        args = []

        if len(self.args) == 0:
            args.append(assg_expr("args", "[]"))
        else:
            for arg in self.args:
                args.append(to_doc(arg))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "WINDOW":
        return self.__class__(*self.args, over=_rebase_clause(self.over, pre))

    def _serialize(self, ctx: SerializationContext) -> None:
        if self.over is not None:
            serialize(self.over, ctx)

        if len(self.args) > 0:
            ctx.newline()
            ctx.write("WINDOW")
            serialize_lines(self.args, ctx)


class WITH(SQLClause):
    args: list[SQLClause]
    recursive: bool
    over: Optional[SQLClause]

    def __init__(
        self,
        *args: SQLClause,
        recursive: bool = False,
        over: Optional[SQLClause] = None,
    ) -> None:
        self.args = list(args)
        self.over = _cast_to_clause_skip_none(over)
        self.recursive = recursive

    def pretty_repr(self, ctx: Optional[QuoteContext]) -> "Doc":
        name = "WITH"
        args = []

        if self.recursive:
            args.append(assg_expr("recursive", to_doc(True)))
        if len(self.args) == 0:
            args.append(assg_expr("args", "[]"))
        else:
            for arg in self.args:
                args.append(to_doc(arg))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over), ex)
        return ex

    def rebase(self, pre: SQLClause) -> "WITH":
        return self.__class__(
            *self.args,
            recursive=self.recursive,
            over=_rebase_clause(self.over, pre),
        )

    def _serialize(self, ctx: SerializationContext) -> None:
        if len(self.args) > 0:
            ctx.write("WITH ")
            if self.recursive and ctx.dialect.has_recursive_annotation:
                ctx.write("RECURSIVE ")

            for i, arg in enumerate(self.args):
                if i > 0:
                    ctx.write(", ")
                    ctx.newline()
                # NOTE: The `WITH` clause inverts order of the alias and the original clause for an `AS` expression
                if isinstance(arg, AS):
                    serialize(arg.name, ctx)
                    if arg.columns is not None:
                        with ctx.parens(space=True):
                            serialize(arg.columns, ctx)
                    ctx.write(" AS ")
                    # NOTE: reassigning `arg` to the parent of the AS clause
                    arg = arg.over

                nested_orig = ctx.nested
                ctx.nested = True
                serialize(arg, ctx)
                ctx.nested = nested_orig
            ctx.newline()

        if self.over is not None:
            serialize(self.over, ctx)


# -----------------------------------------------------------
# utility routines to export
# -----------------------------------------------------------


def qual(*_args: Union[Symbol, str]) -> "ID":
    """Qualify a table/column name, i.e.

    - `qual(schema, table) = ID(schema) >> ID(table)`
    - `qual(table, column) = ID(table) >> ID(column)`, etc.
    """
    args = [S(arg) for arg in _args]
    if len(args) == 2:
        return ID(args[0]) >> ID(args[1])
    elif len(args) == 3:
        return ID(args[0]) >> ID(args[1]) >> ID(args[2])
    else:
        raise NotImplementedError(
            "How do you qualify a SQL identifier with <2, >3 args?"
        )


def alias(curr: Union[SQLClause, Symbol, str], rename: Union[Symbol, str]) -> "AS":
    """Rename an identifier or SQL clause"""
    if not isinstance(curr, SQLClause):
        curr = ID(S(curr))
    return curr >> AS(S(rename))



"""
This module defines types for all the table/column references across 
the FunSQL node objects. Before rendering, the compiler resolves the 
names and types of all references _available_ at a node; which it can 
provide to the nodes downstream. 
"""

from enum import Enum
from typing import Optional, Union, ClassVar, overload

from ..common import S, Symbol
from ..prettier import (
    Doc,
    QuoteContext,
    call_expr,
    assg_expr,
    list_expr,
    annotate_expr,
    to_doc,
)


NEGATIVE_INT = -1000  # placeholder negative value as stand-in for missing index

SQLType = Union["UnitType", "RowType", "BoxType"]


class UnitType(Enum):
    Empty = 1  # placeholder type assigned to a reference while we deduce it
    Scalar = 2  # regular type assigned to a column reference
    Ambiguous = 3  # type assigned to references we are unsure about

    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        return f"{self.name}Type()"


FieldTypeMap = dict["Symbol", SQLType]
HandleTypeMap = dict[int, SQLType]


class RowType:
    """Container for a set of fields mapped to their corresponding types, say
    the column references available at a tabular node.
    """

    fields: "FieldTypeMap"
    group: SQLType

    def __init__(
        self, fields: Optional[FieldTypeMap] = None, group: Optional[SQLType] = None
    ) -> None:
        self.fields = fields if fields is not None else dict()
        self.group = group if group is not None else UnitType.Empty

    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "RowType"
        args = []

        for field, val in self.fields.items():
            args.append(annotate_expr(str(field), to_doc(val)))
        if not self.group == UnitType.Empty:
            args.append(assg_expr("group", to_doc(self.group)))
        return call_expr(name, args)


class BoxType:
    """Type assigned to a Box node (wrapping a tabular node). It tracks the set of
    column references available at the tabular node, and handles.
    """

    name: Symbol
    row: RowType
    handle_map: HandleTypeMap

    def __init__(
        self, name: Symbol, row: "RowType", handle_map: Optional["HandleTypeMap"] = None
    ):
        self.name = name
        self.row = row
        self.handle_map = handle_map if handle_map is not None else dict()

    def is_empty(self) -> bool:
        return self.name.data == "_" and len(self.row.fields) == 0

    def pretty_repr(self, ctx: QuoteContext) -> "Doc":
        name = "BoxType"
        args = []

        args.append(to_doc(self.name))
        for field, val in self.row.fields.items():
            args.append(annotate_expr(str(field), to_doc(val, ctx)))
        if not self.row.group == UnitType.Empty:
            args.append(assg_expr("group", to_doc(self.row.group, ctx)))
        if len(self.handle_map) > 0:
            embedded_handles = list_expr([str(k) for k in self.handle_map])
            args.append(assg_expr("handles_incl", embedded_handles))
            # NOTE: printing the namespace of the handle nodes too makes output string too large
            # for h in sorted(self.handle_map.keys()):
            #     args.append(annotate_expr(str(h), to_doc(self.handle_map[h], ctx)))

        return call_expr(name, args)

    def add_handle(self, handle: int) -> "BoxType":
        # missing handle means a negative int input
        if handle < 0:
            return self
        else:
            return self.__class__(
                self.name, self.row, {**self.handle_map, handle: self.row}
            )


EMPTY_BOX_TYPE = lambda: BoxType(S("_"), RowType())


def intersect(t1: SQLType, t2: SQLType) -> SQLType:
    """Used to deduce the references available at an Append node.

    The union of multiple tables results in a single table, so you'd expect
    only the references available in each of them; so an `intersect` operation.
    """
    if t1 == UnitType.Ambiguous and t2 == UnitType.Ambiguous:
        return UnitType.Ambiguous
    elif t1 == UnitType.Scalar and t2 == UnitType.Scalar:
        return UnitType.Scalar

    elif isinstance(t1, RowType) and isinstance(t2, RowType):
        if t1 == t2:
            return t1
        common_fields: FieldTypeMap = dict()
        for field in set(t1.fields).intersection(set(t2.fields)):
            t = intersect(t1.fields[field], t2.fields[field])
            if not t == UnitType.Empty:
                common_fields[field] = t
        group = intersect(t1.group, t2.group)
        return RowType(common_fields, group)

    elif isinstance(t1, BoxType) and isinstance(t2, BoxType):
        if t1 == t2:
            return t1
        common_handles: HandleTypeMap = dict()
        for key in set(t1.handle_map).intersection(set(t2.handle_map)):
            t = intersect(t1.handle_map[key], t2.handle_map[key])
            if not t == UnitType.Empty:
                common_handles[key] = t
        name = t1.name if t1.name == t2.name else S("union")
        return BoxType(name, intersect(t1.row, t2.row), common_handles)  # type: ignore

    else:
        return UnitType.Empty


def union(t1: SQLType, t2: SQLType) -> SQLType:
    """Used to deduce the box type (read: references available) at a Join
    node, using the box types for the nodes being joined.

    You'd expect the nodes downstream of a Join to access references on either
    of the two sides of the Join, so a `union` makes sense.
    """
    if t1 == UnitType.Empty and t2 == UnitType.Empty:
        return UnitType.Empty
    elif t1 == UnitType.Empty and not t2 == UnitType.Empty:
        return t2
    elif not t1 == UnitType.Empty and t2 == UnitType.Empty:
        return t1
    elif t1 == UnitType.Scalar and t2 == UnitType.Scalar:
        return UnitType.Scalar

    elif isinstance(t1, RowType) and isinstance(t2, RowType):
        combined_fields: FieldTypeMap = dict()
        t1_fields = set(t1.fields)
        t2_fields = set(t2.fields)

        for ref in t1_fields.intersection(t2_fields):
            ref_typ_t1 = t1.fields[ref]
            ref_typ_t2 = t2.fields[ref]
            if isinstance(ref_typ_t1, RowType) and isinstance(ref_typ_t2, RowType):
                combined_fields[ref] = union(ref_typ_t1, ref_typ_t2)
            else:
                combined_fields[ref] = UnitType.Ambiguous
        for ref in t1_fields.difference(t2_fields):
            combined_fields[ref] = t1.fields[ref]
        for ref in t2_fields.difference(t1_fields):
            combined_fields[ref] = t2.fields[ref]

        group = (
            t1.group
            if t2.group == UnitType.Empty
            else t2.group
            if t1.group == UnitType.Empty
            else UnitType.Ambiguous
        )
        return RowType(combined_fields, group)

    elif isinstance(t1, BoxType) and isinstance(t2, BoxType):
        combined_handles: HandleTypeMap = dict()
        t1_handles = set(t1.handle_map)
        t2_handles = set(t2.handle_map)

        for field in t1_handles.intersection(t2_handles):
            combined_handles[field] = UnitType.Ambiguous
        for field in t1_handles.difference(t2_handles):
            combined_handles[field] = t1.handle_map[field]
        for field in t2_handles.difference(t1_handles):
            combined_handles[field] = t2.handle_map[field]
        return BoxType(t1.name, union(t1.row, t2.row), combined_handles)  # type: ignore

    else:
        return UnitType.Ambiguous


def is_subset(t1: SQLType, t2: SQLType) -> bool:
    """Check whether t1 is a subset of t2"""
    if isinstance(t1, RowType) and isinstance(t2, RowType):
        if t1 == t2:
            return True
        for field in t1.fields.keys():
            if field not in t2.fields:
                return False
            if not is_subset(t1.fields[field], t2.fields[field]):
                return False
        return True

    elif isinstance(t1, BoxType) and isinstance(t2, BoxType):
        if t1 == t2:
            return True
        if t1.name != t2.name:
            return False
        if not is_subset(t1.row, t2.row):
            return False
        for field in t1.handle_map.keys():
            if field not in t2.handle_map:
                return False
            if not is_subset(t1.handle_map[field], t2.handle_map[field]):
                return False
        return True

    else:
        # for the non-container SQLType instances, just compare the data types
        return type(t1) == type(t2)



"""
This module implements the `resolve` pass over the annotated SQLNode expression. 
We derive the names and types of references available at all the nodes, by walking 
the query tree, _starting_ at the source. Only the table/column references _available_ 
at a node can be referred by the nodes downstream of it. 
"""

from typing import Any, Optional, Union
from functools import singledispatch

from .types import *
from .annotate import *
from ..common import register_union_type
from ..nodes import *
from ..nodedefs import *

__all__ = ["resolve_toplevel"]


def resolve_toplevel(ctx: AnnotateContext) -> None:
    resolve_boxes(ctx.boxes, ctx)


def resolve_boxes(boxes: list[Box], ctx: AnnotateContext) -> None:
    """Resolve the types for all the nodes that box a tabular node. Note this starts from
    the source nodes (From, CTEs, etc), and propagates the derived namespaces downwards.
    """
    for box in boxes:
        if box.over is not None:
            handle_idx = ctx.get_handle(box.over)
            typ = resolve(box.over, ctx)
            typ = typ.add_handle(handle_idx)
            box.handle = handle_idx
            box.typ = typ


@singledispatch
def resolve(data: Union[TabularNode, As], ctx: AnnotateContext) -> BoxType:
    raise NotImplementedError(f"Don't know how to resolve node of type {type(data)}")


@resolve.register
def _(node: Append, ctx: AnnotateContext) -> BoxType:
    """Iterate over the nodes included in the Union operator. Only
    the references available at all the nodes can be used downstream.
    """
    t = box_type(node.over)
    for arg in node.args:
        t = intersect(t, box_type(arg))
    assert isinstance(t, BoxType)
    return t


@register_union_type(resolve)
def _(node: Union[As, Knot], ctx: AnnotateContext) -> BoxType:
    """As/Knot nodes add an indirection to the column references. For ex, considering
    the query: `From(Table[col1, col2]) >> As(table_alias)`,
    * From(...): has references to col1 and col2
    * As(...): has references to table_alias.col1 and table_alias.col2
    """
    t = box_type(node.over)
    fields: FieldTypeMap = dict()
    fields[node.name] = t.row

    row = RowType(fields)
    return BoxType(node.name, row, t.handle_map)


@resolve.register
def _(node: Define, ctx: AnnotateContext) -> BoxType:
    """Add the column references defined in this node. If the parent node
    has a reference with the same name, it gets overriden.
    """
    t = box_type(node.over)
    fields: FieldTypeMap = dict()
    for f, ft in t.row.fields.items():
        if f not in node.label_map:
            fields[f] = ft
    for f in node.label_map:
        fields[f] = UnitType.Scalar

    row = RowType(fields, group=t.row.group)
    return BoxType(t.name, row, t.handle_map)


@resolve.register
def _(node: FromNothing, ctx: AnnotateContext) -> BoxType:
    return EMPTY_BOX_TYPE()


@resolve.register
def _(node: FromReference, ctx: AnnotateContext) -> BoxType:
    """Check that the parent node had {node.name} as a valid reference. Raises
    an error if not, else copy over the set of references.
    """
    t = box_type(node.over)
    ft = t.row.fields.get(node.name, None)
    if not isinstance(ft, RowType):
        raise ErrRefInvalidTable(name=node.name, path=ctx.get_path(node.over))
    return BoxType(node.name, ft)


@resolve.register
def _(node: FromTable, ctx: AnnotateContext) -> BoxType:
    """Copy over all the column references from the source table."""
    fields: FieldTypeMap = dict()
    for f in node.table.columns:
        fields[f] = UnitType.Scalar
    row = RowType(fields)
    return BoxType(node.table.name, row)


@resolve.register
def _(node: FromValues, ctx: AnnotateContext) -> BoxType:
    """All columns provided by the table of values can be referenced."""
    fields: FieldTypeMap = dict()
    for f in node.source.columns:
        fields[f] = UnitType.Scalar
    row = RowType(fields)
    return BoxType(S("values"), row)


@resolve.register
def _(node: Group, ctx: AnnotateContext) -> BoxType:
    """The columns we are grouping by are available as references, while the ones which
    can be aggregated over are moved to the `group` attribute.
    """
    t = box_type(node.over)
    fields: FieldTypeMap = dict()
    for f in node.label_map:
        fields[f] = UnitType.Scalar
    row = RowType(fields, group=t.row)
    return BoxType(t.name, row)


@register_union_type(resolve)
def _(node: Union[IntBind, Limit, Order, Where, With], ctx: AnnotateContext) -> BoxType:
    """The references available at the parent node can be used here too."""
    return box_type(node.over)


def resolve_knot(node: Box, ctx: AnnotateContext) -> None:
    """The namespace at Knot node is inferred in the following order.
    * Base query namespace -> Knot namespace
    * Knot namespace -> Iterator query namespace
    * Knot namepsace -> IntIterate namespace

    When the IntIterate node is encountered, Knot and Iterator namespaces have
    been resolved once. However, the Knot node, being a union of the base query
    and the iterator query, can only contain references common to both of them.
    Hence, a redo is done.
    """
    assert isinstance(node.over, Knot), "expected a Knot node as parent"
    knot = node.over

    iterator_typ = box_type(knot.iterator)
    while not is_subset(node.typ.row, iterator_typ.row):  # type: ignore
        node.typ = intersect(node.typ, iterator_typ)  # type: ignore
        resolve_boxes(knot.iterator_boxes, ctx)
        iterator_typ = box_type(knot.iterator)
    assert isinstance(node.typ, BoxType)
    node.typ = node.typ.add_handle(node.handle)


@resolve.register
def _(node: IntIterate, ctx: AnnotateContext) -> BoxType:
    """Infer the same references as available at the Knot node"""
    assert isinstance(node.over, Box)
    resolve_knot(node.over, ctx)

    t = box_type(node.over)
    row = t.row.fields[node.iterator_name]
    assert isinstance(row, RowType), "expected a RowType"
    return BoxType(node.name, row)


@resolve.register
def _(node: IntJoin, ctx: AnnotateContext) -> BoxType:
    """All references available on either side of the Join nodes are available."""
    lt = box_type(node.over)
    rt = box_type(node.joinee)
    t: BoxType = union(lt, rt)  # type: ignore
    return t


@resolve.register
def _(node: Partition, ctx: AnnotateContext) -> BoxType:
    """All the references from the parent node are available both as regualar
    columns, and to aggregate over.

    NOTE: This is different from Group node, where you can't access columns not
    used to group, only aggregate over them.
    """
    t = box_type(node.over)
    row = RowType(t.row.fields, group=t.row)
    return BoxType(t.name, row, t.handle_map)


@resolve.register
def _(node: Select, ctx: AnnotateContext) -> BoxType:
    """All the references included in the select node are available downstream."""
    t = box_type(node.over)
    fields: FieldTypeMap = dict()
    for f in node.label_map:
        fields[f] = UnitType.Scalar
    row = RowType(fields)
    return BoxType(t.name, row)



"""
The `link` step resolves the columns provided by each tabular node. 

* We start with the leaf `Box` node, which typically wraps a Select node, and add 
the columns used in the output to its `refs` list, that is the list of references it 
needs to provide. 
* Then, we propagate upwards. 
  * Each tabular node wrapped in a `Box` node reads the `refs` that its box needs to provide. 
  * It further adds the references it makes use of, for ex, arguments to a `Fun` node, or the 
  condition in a `Where` node. 
  * It assigns each of these references to one of its parent `Box` nodes. 
  * These box nodes further link references through their parent tabular nodes. 
"""

from typing import Any, Optional, Union, TypeVar, Literal
from functools import singledispatch

from .types import *
from .annotate import *
from ..common import register_union_type
from ..nodes import *
from ..nodedefs import *

__all__ = ["link_toplevel"]


# -----------------------------------------------------------
# utilities to help link references
# -----------------------------------------------------------


def gather(node: Union[None, SQLNode, list[SQLNode]], refs: list[SQLNode]) -> None:
    """Gather node references in the `refs` container; that are used by the descendants of
    the box node which owns the container.
    """
    if isinstance(node, list):
        for n in node:
            assert isinstance(n, SQLNode), f"expected a SQLNode, got {type(n)}"
            gather(n, refs)
    elif isinstance(node, (As, Box, Sort)):
        gather(node.over, refs)
    elif isinstance(node, IntBind):
        gather(node.over, refs)
        gather(node.args, refs)
        node.owned = True
    elif isinstance(node, Fun):
        gather(node.args, refs)
    elif isinstance(node, (Agg, Get, HandleBound, NameBound)):
        refs.append(node)
    else:
        pass


def validate(t: SQLType, ref: SQLNode, ctx: AnnotateContext) -> None:
    """validate references"""
    if isinstance(t, RowType):
        validate_rowtype(t, ref, ctx)
    elif isinstance(t, BoxType):
        validate_boxtype(t, ref, ctx)
    else:
        raise NotImplementedError(f"validate method not implemented for: {type(t)}")


def validate_boxtype(t: BoxType, ref: SQLNode, ctx: AnnotateContext) -> None:
    if isinstance(ref, HandleBound):
        handle = ref.handle  # index to the tabular node used as handle
        over = ref.over  # TODO: must this be a Get node?
        if handle in t.handle_map:
            ht = t.handle_map[handle]
            if ht == UnitType.Ambiguous:
                raise ErrRefAmbiguousHandle(path=ctx.get_path(ref))
            else:
                validate(ht, over, ctx)
        else:
            raise ErrRefUndefinedHandle(path=ctx.get_path(ref))

    else:
        validate(t.row, ref, ctx)


def validate_rowtype(t: RowType, ref: SQLNode, ctx: AnnotateContext) -> None:
    """verify that the type container has an entry for the node referenced"""
    while isinstance(ref, NameBound):
        name = ref.name
        ft = t.fields.get(name, UnitType.Empty)
        if not isinstance(ft, RowType):
            if ft == UnitType.Empty:
                raise ErrRefUndefinedName(name=name, path=ctx.get_path(ref))
            elif ft == UnitType.Scalar:
                raise ErrRefUnexpectedScalarType(name=name, path=ctx.get_path(ref))
            elif ft == UnitType.Ambiguous:
                raise ErrRefAmbiguousName(name=name, path=ctx.get_path(ref))
            else:
                raise Exception(f"unexpected field type: {type(ft)}")
        t = ft
        ref = ref.over

    if isinstance(ref, Get) and ref.over is None:
        name = ref._name
        ft = t.fields.get(name, UnitType.Empty)
        if not ft == UnitType.Scalar:
            if ft == UnitType.Empty:
                raise ErrRefUndefinedName(name=name, path=ctx.get_path(ref))
            elif ft == UnitType.Ambiguous:
                raise ErrRefAmbiguousName(name=name, path=ctx.get_path(ref))
            elif isinstance(ft, RowType):
                raise ErrRefUnexpectedRowType(name=name, path=ctx.get_path(ref))
            else:
                raise Exception(f"unexpected field type: {type(ft)}")

    elif isinstance(ref, Agg) and ref.over is None:
        name = ref._name
        if not isinstance(t.group, RowType):
            if t.group == UnitType.Empty:
                raise ErrRefUnexpectedAgg(name=name, path=ctx.get_path(ref))
            elif t.group == UnitType.Ambiguous:
                raise ErrRefAmbiguousAgg(name=name, path=ctx.get_path(ref))
            else:
                raise Exception(f"unexpected group type: {type(t.group)}")

    else:
        raise Exception(f"unexpected reference: {type(ref)}")


def gather_n_validate(
    node: Union[SQLNode, list[SQLNode]],
    parent_refs: list[SQLNode],
    parent_t: BoxType,
    ctx: AnnotateContext,
) -> None:
    """Gather all the columns `node` requires and add them to the list of references
    the parent box node needs to provide. Then, validate that they indeed exist in the
    parent box namespace.
    """
    start_at = len(parent_refs)
    gather(node, parent_refs)
    for ref in parent_refs[start_at:]:
        validate(parent_t, ref, ctx)


T = TypeVar("T", BoxType, RowType)


def route(lt: T, rt: T, ref: SQLNode) -> Literal[-1, 1]:
    """For all the references at a Join node, determine if a reference was sourced
    from the left side node of the join or the right.

    Returns:
      * -1 if the reference was sourced from the left side node
      * 1 if the reference was sourced from the right side node
    """

    if isinstance(lt, BoxType) and isinstance(rt, BoxType):
        if isinstance(ref, HandleBound):
            typ = lt.handle_map.get(ref.handle, UnitType.Empty)
            return 1 if typ == UnitType.Empty else -1
        else:
            return route(lt.row, rt.row, ref)

    elif isinstance(lt, RowType) and isinstance(rt, RowType):
        while isinstance(ref, NameBound):
            lt_p = lt.fields.get(ref.name, UnitType.Empty)
            if lt_p == UnitType.Empty:
                return 1
            rt_p = rt.fields.get(ref.name, UnitType.Empty)
            if rt_p == UnitType.Empty:
                return -1
            assert isinstance(lt_p, RowType) and isinstance(rt_p, RowType)
            lt = lt_p
            rt = rt_p
            ref = ref.over

        if isinstance(ref, Get):
            return -1 if ref._name in lt.fields else 1
        elif isinstance(ref, Agg):
            return -1 if isinstance(lt.group, RowType) else 1
        else:
            raise Exception(f"unexpected ref node of type: {type(ref)}")

    else:
        raise Exception(f"unexpected nodes of type: {type(lt)}")


# -----------------------------------------------------------
# linking implemented
# -----------------------------------------------------------


def link_toplevel(ctx: AnnotateContext) -> None:
    # `leaf_box` boxes the query node, that is the one corresponding to the query output
    leaf_box = ctx.boxes[-1]
    for f, ft in leaf_box.typ.row.fields.items():
        if ft == UnitType.Scalar:
            leaf_box.refs.append(Get(f))

    # now, we start again from the `leaf_box`, and link references through all the box nodes
    link_boxes(list(reversed(ctx.boxes)), ctx)


def link_boxes(boxes: list[Box], ctx: AnnotateContext) -> None:
    for box in boxes:
        if box.over is not None:
            refs_p: list[SQLNode] = []
            for ref in box.refs:
                if isinstance(ref, HandleBound) and ref.handle == box.handle:
                    refs_p.append(ref.over)
                else:
                    refs_p.append(ref)
            link(box.over, refs_p, ctx)


def check_box(node: Optional[SQLNode]) -> Box:
    assert isinstance(node, Box), f"expected node of type: Box, got: {type(node)}"
    return node


@singledispatch
def link(node: SQLNode, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    raise NotImplementedError(f"link method not implemented for: {type(node)}")


@link.register
def _(node: Append, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    box = check_box(node.over)
    box.refs.extend(refs)
    for arg in node.args:
        assert isinstance(arg, Box), "expected `args` of type: Row"
        box = arg
        box.refs.extend(refs)


@link.register
def _(node: As, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    box = check_box(node.over)
    for ref in refs:
        if isinstance(ref, NameBound):
            assert ref.name == node.name
            box.refs.append(ref.over)
        elif isinstance(ref, HandleBound):
            box.refs.append(ref)
        else:
            raise Exception(f"unexpected ref node of type: {type(ref)}")


@link.register
def _(node: Define, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    box = check_box(node.over)
    seen: set[Symbol] = set()
    for ref in refs:
        if isinstance(ref, Get) and ref.over is None and ref._name in node.label_map:
            if ref._name not in seen:
                seen.add(ref._name)
                col = node.args[node.label_map[ref._name]]
                gather_n_validate(col, box.refs, box.typ, ctx)
        else:
            box.refs.append(ref)


@register_union_type(link)
def _(
    node: Union[FromNothing, FromTable, FromValues, FromReference],
    refs: list[SQLNode],
    ctx: AnnotateContext,
) -> None:
    if isinstance(node, FromReference):
        box = check_box(node.over)
        for ref in refs:
            box.refs.append(NameBound(over=ref, name=node.name))


@link.register
def _(node: Group, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    box = check_box(node.over)
    gather_n_validate(node.by, box.refs, box.typ, ctx)
    for ref in refs:
        if isinstance(ref, Agg) and ref.over is None:
            gather_n_validate(ref.args, box.refs, box.typ, ctx)
            if ref.filter_ is not None:
                gather_n_validate(ref.filter_, box.refs, box.typ, ctx)


@register_union_type(link)
def _(node: Union[Limit, With], refs: list[SQLNode], ctx: AnnotateContext) -> None:
    box = check_box(node.over)
    box.refs.extend(refs)


@link.register
def _(node: IntBind, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    if not node.owned:
        gather_n_validate(node.args, [], EMPTY_BOX_TYPE(), ctx)
    box = check_box(node.over)
    box.refs.extend(refs)


@link.register
def _(node: IntIterate, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    box = check_box(node.over)
    for ref in refs:
        box.refs.append(NameBound(over=ref, name=node.iterator_name))


@link.register
def _(node: IntJoin, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    lbox = check_box(node.over)
    rbox = check_box(node.joinee)
    lrefs, rrefs = [], []

    for ref in refs:
        turn = route(lbox.typ, rbox.typ, ref)
        if turn < 0:
            lrefs.append(ref)
        else:
            rrefs.append(ref)
    if len(rrefs) != 0:
        node.skip = False
    if node.skip:
        lbox.refs.extend(lrefs)
        return

    # gather any lateral references (if the right side contains a Bind node)
    gather_n_validate(node.joinee, node.lateral, lbox.typ, ctx)
    lbox.refs.extend(node.lateral)

    refs_p = []
    gather_n_validate(node.on, refs_p, node.box.typ, ctx)
    for ref in refs_p:
        turn = route(lbox.typ, rbox.typ, ref)
        if turn < 0:
            lbox.refs.append(ref)
        else:
            rbox.refs.append(ref)
    lbox.refs.extend(lrefs)
    rbox.refs.extend(rrefs)


@link.register
def _(node: Knot, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    box = check_box(node.over)
    iterator_box = check_box(node.iterator)

    refs = []
    seen = set()
    while True:
        repeat = False
        for ref in node.box.refs:
            assert isinstance(ref, NameBound), "expected `ref` of type: NameBound"
            assert ref.name == node.name
            if ref.over not in seen:
                refs.append(ref)
                seen.add(ref.over)
                repeat = True

        node.box.refs.clear()
        node.box.refs.extend(refs)
        if not repeat:
            break
        for ibox in node.iterator_boxes:
            ibox.refs.clear()
        iterator_box.refs.extend(refs)
        link_boxes(list(reversed(node.iterator_boxes)), ctx)

    for ref in refs:
        assert isinstance(ref, NameBound), "expected `ref` of type: NameBound"
        box.refs.append(ref.over)


@link.register
def _(node: Order, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    box = check_box(node.over)
    box.refs.extend(refs)
    gather_n_validate(node.by, box.refs, box.typ, ctx)


@link.register
def _(node: Partition, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    box = check_box(node.over)
    for ref in refs:
        if isinstance(ref, Agg) and ref.over is None:
            gather_n_validate(ref.args, box.refs, box.typ, ctx)
            if ref.filter_ is not None:
                gather_n_validate(ref.filter_, box.refs, box.typ, ctx)
        else:
            box.refs.append(ref)
    gather_n_validate(node.by, box.refs, box.typ, ctx)
    gather_n_validate(node.order_by, box.refs, box.typ, ctx)


@link.register
def _(node: Select, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    """Select only passes through references that are present in its `args` attribute,
    so the input `refs` don't need to be used.
    """
    box = check_box(node.over)
    gather_n_validate(node.args, box.refs, box.typ, ctx)


@link.register
def _(node: Where, refs: list[SQLNode], ctx: AnnotateContext) -> None:
    box = check_box(node.over)
    box.refs.extend(refs)
    gather_n_validate(node.condition, box.refs, box.typ, ctx)



import datetime
import io
from contextlib import contextmanager
from functools import singledispatch
from typing import Any

from ..common import Symbol
from ..sqlcontext import SQLDialect


class SQLString:
    query: str
    variables: list[Symbol]

    def __init__(self, query: str, variables: list[Symbol]) -> None:
        self.query = query
        self.variables = variables

    def __repr__(self) -> str:
        query = f"query: \n{self.query}"
        if len(self.variables) == 0:
            return query
        else:
            variables = f"vars: {self.variables}"
            return f"{query}\n\n{variables}"


class SerializationContext:
    dialect: SQLDialect
    buffer: io.StringIO
    level: int
    nested: bool
    variables: list[Symbol]

    def __init__(self, dialect: SQLDialect) -> None:
        self.dialect = dialect
        self.buffer = io.StringIO()
        self.level = 0
        self.nested = False
        self.variables = []

    def write(self, data: str) -> None:
        self.buffer.write(data)

    def newline(self) -> None:
        self.buffer.write("\n")
        for _ in range(self.level):
            self.buffer.write("  ")

    @contextmanager
    def parens(self, space: bool = False):
        self.write(" (" if space else "(")
        yield
        self.write(") " if space else ")")

    def render(self) -> SQLString:
        return SQLString(self.buffer.getvalue(), self.variables.copy())


@singledispatch
def serialize(data: Any, ctx: SerializationContext) -> None:
    if hasattr(data, "_serialize"):
        data._serialize(ctx)
    else:
        raise NotImplementedError(
            f"Don't know how to serialize data of type: {type(data)}"
        )


@serialize.register
def _(data: int, ctx: SerializationContext) -> None:
    ctx.write(str(data))


@serialize.register
def _(data: float, ctx: SerializationContext) -> None:
    ctx.write(str(data))


@serialize.register
def _(data: bool, ctx: SerializationContext) -> None:
    if ctx.dialect.has_bool_literals:
        ctx.write("TRUE" if data else "FALSE")
    else:
        ctx.write("(1 = 1)" if data else "(1 = 0)")


@serialize.register
def _(data: datetime.datetime, ctx: SerializationContext) -> None:
    # TODO: Should we raise an exception when dealing with date/time values for the SQLite dialect?
    ctx.write(f"TIMESTAMP '{data.isoformat()}'")


@serialize.register
def _(data: datetime.date, ctx: SerializationContext) -> None:
    ctx.write(f"DATE '{data.strftime('%Y-%m-%d')}'")


@serialize.register
def _(data: datetime.time, ctx: SerializationContext) -> None:
    """Python resolves time values to count of hours, mins, seconds and microseconds;
    we serialize it to the resolution needed.
    """
    format_str = "%H:%M:%S" if data.microsecond == 0 else "%H:%M:%S.%f"
    ctx.write(f"TIME '{data.strftime(format_str)}'")


@serialize.register
def _(data: datetime.timedelta, ctx: SerializationContext) -> None:
    """Python resolves timedelta values to count of days, seconds, microseconds;
    we serialize it to the resolution needed.
    """
    ctx.write(f"INTERVAL ")
    if data.microseconds == 0:
        if data.seconds == 0:
            ctx.write(f"'{data.days}' DAY")
        else:
            seconds = data.days * 86400 + data.seconds
            ctx.write(f"'{seconds}' SECOND")
    else:
        ctx.write(f"'{data.total_seconds():.6f}' SECOND")


@serialize.register(type(None))
def _(data: None, ctx: SerializationContext) -> None:
    ctx.write("NULL")


@serialize.register
def _(data: str, ctx: SerializationContext) -> None:
    if "'" in data:
        data = data.replace("'", "''")  # tis the postgres way
    ctx.write(f"'{data}'")


@serialize.register
def _(data: Symbol, ctx: SerializationContext) -> None:
    s = str(data)
    lq, rq = ctx.dialect.id_quotes
    if rq in s:
        s = s.replace(rq, rq + rq)
    ctx.write(f"{lq}{s}{rq}")



"""
This module implements the `annotation` pass over the SQLNode expression, which 
rewrites the node graph to refactor some of the nodes (Get, Iterate, etc.) and 
boxes all of the tabular nodes. 
"""

from contextlib import contextmanager
from functools import singledispatch
from typing import Any, Optional, Mapping, Union, overload

from ..common import Symbol, register_union_type
from ..nodedefs import *
from ..nodes import *
from ..prettier import (
    Doc,
    QuoteContext,
    call_expr,
    assg_expr,
    list_expr,
    pipe_expr,
    to_doc,
)
from ..sqlcontext import SQLCatalog, SQLTable
from .types import *


__all__ = [
    "Box",
    "NameBound",
    "HandleBound",
    "FromNothing",
    "FromTable",
    "FromReference",
    "FromValues",
    "IntBind",
    "IntIterate",
    "Knot",
    "IntJoin",
    "AnnotateContext",
    "box_type",
    "annotate",
]


def _rebase_node(curr: Optional[SQLNode], pre: SQLNode) -> SQLNode:
    """shorthand for a common pattern"""
    return pre if curr is None else curr.rebase(pre)


# -----------------------------------------------------------
# Auxiliary node definitions
# -----------------------------------------------------------


class Box(TabularNode):
    """
    Represents a SQL query node with SELECT args undetermined. A box
    node wraps a regular SQLNode object, and is a container for information
    like, the column/table references available at the node which can be
    accessed downstream, pointers to other nodes referred by it, etc.

    Attributes:
        typ: encapsulates the list of references available at this node
        handle: Filled in during the `resolve` step, if the node is used as a handle
        refs: list of SQLNode objects referred by this node
        over: original SQLNode object boxed by this node
    """

    typ: BoxType
    handle: int
    refs: list[SQLNode]
    over: Optional[SQLNode]

    def __init__(
        self,
        typ: Optional[BoxType] = None,
        handle: int = NEGATIVE_INT,
        refs: Optional[list[SQLNode]] = None,
        over: Optional[SQLNode] = None,
    ) -> None:
        super().__init__()
        self.typ = typ if typ is not None else EMPTY_BOX_TYPE()
        self.handle = handle
        self.refs = refs if refs is not None else []
        self.over = over

    def rebase(self, pre: TabularNode) -> "Box":
        return Box(self.typ, self.handle, self.refs, _rebase_node(self.over, pre))

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "Box"
        args = []

        if ctx.limit:
            args.append("...")
        else:
            if not self.typ.is_empty():
                args.append(assg_expr("type", to_doc(self.typ)))
            if self.handle >= 0:
                args.append(assg_expr("handle", str(self.handle)))
            if len(self.refs) > 0:
                args.append(
                    assg_expr("refs", list_expr([to_doc(r, ctx) for r in self.refs]))
                )

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex


def box_type(node: Union[None, SQLNode, Box]) -> BoxType:
    return node.typ if isinstance(node, Box) else EMPTY_BOX_TYPE()


class NameBound(SQLNode):
    """
    Represents a hierarchical Get node, obtained by inverting a regular
    Get node. The node `Get.a.b` is transformed as:

    `Get(over = Get(S.a), name = S.b) => NameBound(over = Get(S.b), name = S.a)`.

    Another example: `Get.A.B.col` (which is equivalent to `Get.A >> Get.B >> Get.col`),
    gets transformed to, `Get.col >> Namebound(S.B) >> NameBound(S.A)`.
    """

    over: SQLNode
    name: Symbol

    def __init__(self, over: SQLNode, name: Symbol) -> None:
        super().__init__()
        self.over = over
        self.name = name

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "NameBound"
        args = []

        args.append(assg_expr("over", to_doc(self.over, ctx)))
        args.append(assg_expr("name", str(self.name)))
        return call_expr(name, args)


class HandleBound(SQLNode):
    """Represents an identifier bound to a regular SQL node, obtained by inverting a
    bound reference. The node `q.a` is tranformed as:

    `Get(over = q, name = S.a) => HandleBound(over = Get(S.a), handle = get_handle(q))`
    """

    over: SQLNode
    handle: int

    def __init__(self, over: SQLNode, handle: int) -> None:
        super().__init__()
        self.over = over
        self.handle = handle

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "HandleBound"
        args = []

        args.append(assg_expr("over", to_doc(self.over, ctx)))
        args.append(assg_expr("handle", str(self.handle)))
        return call_expr(name, args)


# A From node is specialized to one of:
# - FromNothing: stand-in for a From node without a reference
# - FromTable: stand-in for a From node with a table as source
# - FromReference: stand-in for a From node referencing a source through an alias
# - FromValues: stand-in for a From node with a list of values as source


class FromNothing(TabularNode):
    def __init__(self) -> None:
        super().__init__()

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        return call_expr("FromNothing", [])


class FromTable(TabularNode):
    table: SQLTable

    def __init__(self, table: SQLTable) -> None:
        super().__init__()
        self.table = table

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "FromTable"
        args = []

        alias = ctx.vars_.get(self.table, None)
        if alias is not None:
            args.append(str(alias))
        else:
            args.append(to_doc(self.table, QuoteContext(limit=True)))
        return call_expr(name, args)


class FromReference(TabularNode):
    over: SQLNode
    name: Symbol

    def __init__(self, over: SQLNode, name: Symbol) -> None:
        super().__init__()
        self.over = over
        self.name = name

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "FromReference"
        args = []

        args.append(assg_expr("name", str(self.name)))
        args.append(assg_expr("over", to_doc(self.over, ctx)))
        return call_expr(name, args)


class FromValues(TabularNode):
    source: ValuesTable

    def __init__(self, source: ValuesTable) -> None:
        super().__init__()
        self.source = source

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "FromValues"
        args = []

        for col in self.source.columns:
            args.append(assg_expr(str(col), "[...]"))
        return call_expr(name, args)


class IntBind(SQLNode):
    """Substitute for a Bind node after the annotate step."""

    args: list[SQLNode]
    over: Optional[SQLNode]
    label_map: dict[Symbol, int]
    owned: bool  # is the outer query for this node resolved?

    def __init__(
        self,
        args: list[SQLNode],
        owned: bool = False,
        over: Optional[SQLNode] = None,
    ) -> None:
        super().__init__()
        self.args = list(args)
        self.owned = owned
        self.over = over
        self.label_map = populate_label_map(self, self.args)

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "IntBind"
        args = []

        if len(self.args) == 0:
            args.append(assg_expr("args", to_doc("[]")))
        else:
            for arg in self.args:
                args.append(to_doc(arg, ctx))
        args.append(assg_expr("owned", str(self.owned)))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex

    def rebase(self, pre: SQLNode) -> SQLNode:
        return IntBind(
            args=self.args,
            owned=self.owned,
            over=_rebase_node(self.over, pre),
        )


class Knot(TabularNode):
    """An Iterate node (recursive UNION ALL) is split into a Knot and an IntIterate node."""

    name: Symbol  # label of the iterator node
    box: Box  # reference to the Box node wrapping this
    iterator: SQLNode
    iterator_boxes: list[Box]
    over: Optional[SQLNode]

    def __init__(
        self,
        iterator: SQLNode,
        box: Box,
        name: Optional[Symbol] = None,
        iterator_boxes: Optional[list[Box]] = None,
        over: Optional[SQLNode] = None,
    ):
        super().__init__()
        self.iterator = iterator
        self.box = box
        self.name = label(iterator) if name is None else name
        self.iterator_boxes = [] if iterator_boxes is None else iterator_boxes
        self.over = over

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "Knot"
        args = []

        args.append(assg_expr("name", str(self.name)))
        if ctx.limit:
            args.append("...")
        else:
            args.append(assg_expr("iterator", to_doc(self.iterator, ctx)))
            boxes = list_expr([to_doc(b, ctx) for b in self.iterator_boxes])
            args.append(assg_expr("iterator_boxes", boxes))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex

    def rebase(self, pre: SQLNode) -> SQLNode:
        return Knot(
            iterator=self.iterator,
            box=self.box,
            name=self.name,
            iterator_boxes=self.iterator_boxes,
            over=_rebase_node(self.over, pre),
        )


class IntIterate(TabularNode):
    """An Iterate node (recursive UNION ALL) is split into a Knot and an IntIterate node."""

    name: Symbol  # label of the parent to Iterate node
    iterator_name: Symbol  # label of the iterator
    over: Optional[SQLNode]

    def __init__(
        self, name: Symbol, iterator_name: Symbol, over: Optional[SQLNode] = None
    ):
        super().__init__()
        self.name = name
        self.iterator_name = iterator_name
        self.over = over

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "IntIterate"
        args = []
        args.append(assg_expr("name", str(self.name)))
        args.append(assg_expr("iterator_name", str(self.iterator_name)))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex

    def rebase(self, pre: SQLNode) -> SQLNode:
        return IntIterate(
            name=self.name,
            iterator_name=self.iterator_name,
            over=_rebase_node(self.over, pre),
        )


class IntJoin(TabularNode):
    """Join nodes are replaced with `IntJoin` nodes, since we also need to track
    the `box` and `lateral` attributes.
    """

    joinee: SQLNode
    on: SQLNode
    left: bool
    right: bool
    skip: bool
    box: Box
    lateral: list[SQLNode]
    over: Optional[SQLNode]

    def __init__(
        self,
        joinee: SQLNode,
        on: SQLNode,
        left: bool = False,
        right: bool = False,
        skip: bool = False,
        box: Optional[Box] = None,
        lateral: Optional[list[SQLNode]] = None,
        over: Optional[SQLNode] = None,
    ):
        super().__init__()
        self.joinee = joinee
        self.on = on
        self.left = left
        self.right = right
        self.skip = skip
        self.box = box if box is not None else Box()
        self.lateral = [] if lateral is None else lateral
        self.over = over

    @check_repr_context
    def pretty_repr(self, ctx: QuoteContext) -> Doc:
        name = "IntJoin"
        args = []

        if ctx.limit:
            args.append("...")
        else:
            args.append(to_doc(self.joinee, ctx))
            args.append(to_doc(self.on, ctx))
            if self.left:
                args.append(assg_expr("left", str(self.left)))
            if self.right:
                args.append(assg_expr("right", str(self.right)))
            if self.skip:
                args.append(assg_expr("skip", str(self.skip)))
            lateral_args = list_expr([to_doc(x, ctx) for x in self.lateral])
            if len(self.lateral) > 0:
                args.append(assg_expr("lateral", lateral_args))

        ex = call_expr(name, args)
        if self.over is not None:
            ex = pipe_expr(to_doc(self.over, ctx), ex)
        return ex

    def rebase(self, pre: SQLNode) -> SQLNode:
        return IntJoin(
            joinee=self.joinee,
            on=self.on,
            left=self.left,
            right=self.right,
            skip=self.skip,
            box=self.box,
            lateral=self.lateral,
            over=_rebase_node(self.over, pre),
        )


# -----------------------------------------------------------
# label method implemented for auxiliary nodes
# -----------------------------------------------------------


@register_union_type(label)
def _(node: Union[IntJoin, IntBind, HandleBound, NameBound]) -> Symbol:
    return label(node.over)


@register_union_type(label)
def _(node: Union[IntIterate, Knot]) -> Symbol:
    return node.name


@label.register
def _(node: Box) -> Symbol:
    # during the annotation phase, Box nodes are not resolved yet so defer to the parent
    return label(node.over) if node.typ.is_empty() else node.typ.name


# -----------------------------------------------------------
# Implements the Annotation context and annotate methods for
# each node
# -----------------------------------------------------------


class PathMap:
    """Maps a node in the annotated graph to a path in the original node graph,
    so errors can be highlighted.

    Attributes:
        - paths: keeps a list of all nodes in the original graph as they are annotated
        - origins: tracks an index to the original node for each annotated/box node
    """

    paths: list[tuple[SQLNode, int]]
    origins: dict[Any, int]

    def __init__(self) -> None:
        self.paths = []
        self.origins = dict()

    def get_path(self, idx_or_node: Union[int, SQLNode]) -> list[SQLNode]:
        idx = (
            idx_or_node
            if isinstance(idx_or_node, int)
            else self.origins.get(idx_or_node, NEGATIVE_INT)
        )
        path = []
        # recursively fetch all nodes till the root
        while idx >= 0:
            node, idx = self.paths[idx]
            path.append(node)
        return path


class AnnotateContext:
    """Container for the data collected during the annotation step

    Attributes:
        - catalog: SQL context during the compile pass
        - path_map: PathMap
        - current_path: stack like variable, tracks all nodes from the root to the current node
        - handles: keeps an index of all original nodes used for bound references
        - boxes: list of all box nodes created during the annotate pass
        - cte_nodes: list of all CTE nodes encountered during the annotate pass
    """

    catalog: SQLCatalog
    path_map: PathMap
    current_path: list[int]
    handles: dict[SQLNode, int]
    boxes: list[Box]
    cte_nodes: dict[Symbol, SQLNode]

    def __init__(self, catalog: SQLCatalog) -> None:
        self.catalog = catalog
        self.path_map = PathMap()
        self.current_path = [NEGATIVE_INT]
        self.handles = dict()
        self.boxes = []
        self.cte_nodes = dict()

    @contextmanager
    def extend_cte_nodes(self, more_nodes: Mapping[Symbol, SQLNode]):
        """Context manager for extending the container of cte_nodes with new labels,
        and restoring the previous value on exit.
        """
        old_nodes = self.cte_nodes
        self.cte_nodes = {**old_nodes, **more_nodes}
        yield
        self.cte_nodes = old_nodes

    @contextmanager
    def extend_path(self, node: SQLNode):
        """Context manager for growing the current path when annotating a new node,
        and shrinking it back after.
        """
        self.path_map.paths.append((node, self.current_path[-1]))
        self.current_path.append(len(self.path_map.paths) - 1)
        yield
        self.current_path.pop()

    def mark_origin(self, node: SQLNode) -> None:
        self.path_map.origins[node] = self.current_path[-1]

    def get_path(self, node: Optional[SQLNode] = None) -> list[SQLNode]:
        dest = self.current_path[-1] if node is None else node
        return self.path_map.get_path(dest)

    def make_handle(self, node: SQLNode) -> int:
        """When encountering a node used as a bound reference, i.e. `node >> Get.X`, give it
        an id so the corresponding reference can track it.
        """
        if node not in self.handles:
            self.handles[node] = len(self.handles)
        return self.handles[node]

    def get_handle(self, node: Optional[SQLNode]) -> int:
        """
        Args:
            A boxed node
        Returns:
            If the original node was used as a handle, return its index in the handles list.
        """
        handle = NEGATIVE_INT
        if node is not None:
            idx = self.path_map.origins.get(node, NEGATIVE_INT)
            if idx >= 0:
                origin_node = self.path_map.paths[idx][0]
                handle = self.handles.get(origin_node, NEGATIVE_INT)
        return handle


# -----------------------------------------------------------
# Rewriting the node graph
# -----------------------------------------------------------


@overload
def annotate(node: None, ctx: AnnotateContext) -> Box:
    ...


@overload
def annotate(node: SQLNode, ctx: AnnotateContext) -> Box:
    ...


@overload
def annotate(node: list[SQLNode], ctx: AnnotateContext) -> list[Box]:
    ...


def annotate(
    node: Union[None, SQLNode, list[SQLNode]], ctx: AnnotateContext
) -> Union[Box, list[Box]]:
    """Annotate a tabular node, by delegating to the type specific annotation
    routine first and then wrapping the output in a box node.
    """
    if isinstance(node, list):
        assert all(isinstance(x, SQLNode) for x in node), "not all nodes are SQLNodes"
        return [annotate(x, ctx) for x in node]

    elif node is None:
        box = Box(over=None)
        ctx.boxes.append(box)
        ctx.mark_origin(box)
        return box

    else:
        # this isn't necessary, just to help me debug errors early
        if not isinstance(node, (TabularNode, As, Bind)):
            raise ErrIllFormed(path=ctx.get_path())
        with ctx.extend_path(node):
            node_p = annotate_node(node, ctx)
            ctx.mark_origin(node_p)

            box = Box(over=node_p)
            ctx.boxes.append(box)
            ctx.mark_origin(box)

            if isinstance(node_p, IntJoin):
                node_p.box = box
        return box


@overload
def annotate_scalar(node: None, ctx: AnnotateContext) -> None:
    ...


@overload
def annotate_scalar(node: SQLNode, ctx: AnnotateContext) -> SQLNode:
    ...


@overload
def annotate_scalar(node: list[SQLNode], ctx: AnnotateContext) -> list[SQLNode]:
    ...


def annotate_scalar(
    node: Union[SQLNode, None, list[SQLNode]], ctx: AnnotateContext
) -> Union[SQLNode, None, list[SQLNode]]:
    """Annotate a scalar node, by delegating to the type specific annotation routine.

    NOTE: generic typing using TypeVar refuses to cooperate, so we have to take the
    longer `overload` route to type the function correctly.
    """
    if isinstance(node, list):
        assert all(isinstance(x, SQLNode) for x in node), "not all nodes are SQLNodes"
        return [annotate_scalar(x, ctx) for x in node]

    elif node is None:
        return None

    elif isinstance(node, TabularNode):
        return annotate(node, ctx)

    else:
        with ctx.extend_path(node):
            node_p = annotate_node_scalar(node, ctx)
            ctx.mark_origin(node_p)
            return node_p


@singledispatch
def annotate_node(node: SQLNode, ctx: AnnotateContext) -> SQLNode:
    raise NotImplementedError(
        f"annotate_node is not implemented for type: {type(node)}"
    )


@singledispatch
def annotate_node_scalar(node: SQLNode, ctx: AnnotateContext) -> SQLNode:
    raise NotImplementedError(
        f"annotate_node_scalar is not implemented for type: {type(node)}"
    )


def rebind(
    parent: Optional[SQLNode], node_isolated: SQLNode, ctx: AnnotateContext
) -> SQLNode:
    """Traverse up the hierarchy and reverse the Get/Agg nodes. Takes as input
    a scalar node split into an isolated copy of it, and its parent node.

    Args:
        parent: The parent node
        node_isolated: The node isolated (i.e. with parent set to None)
        ctx: Annotation context

    Returns:
        The rebinded node
    """

    while isinstance(parent, Get):
        ctx.mark_origin(node_isolated)
        node_isolated = NameBound(over=node_isolated, name=parent._name)
        parent = parent.over
    if parent is not None:
        handle = ctx.make_handle(parent)
        ctx.mark_origin(node_isolated)
        node_isolated = HandleBound(over=node_isolated, handle=handle)
    return node_isolated


@annotate_node_scalar.register
def _(node: Agg, ctx: AnnotateContext) -> SQLNode:
    # TODO: how critical is it to annotate these attributes in order? For each type of node?
    args_p = annotate_scalar(node.args, ctx)
    filter_p = annotate_scalar(node.filter_, ctx)
    node_p = Agg(node._name, *args_p, distinct=node.distinct, filter_=filter_p)
    return rebind(node.over, node_p, ctx)


@annotate_node.register
def _(node: Append, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate(node.over, ctx)
    args_p = annotate(node.args, ctx)
    return Append(*args_p, over=over_p)


@annotate_node.register
def _(node: As, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate(node.over, ctx)
    return As(name=node.name, over=over_p)


@annotate_node_scalar.register
def _(node: As, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate_scalar(node.over, ctx)
    return As(name=node.name, over=over_p)


@annotate_node.register
@annotate_node_scalar.register
def _(node: Bind, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate(node.over, ctx)
    args_p = annotate_scalar(node.args, ctx)
    return IntBind(args=args_p, over=over_p)


@annotate_node.register
def _(node: Define, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate(node.over, ctx)
    args_p = annotate_scalar(node.args, ctx)
    return Define(*args_p, over=over_p)


@annotate_node.register
def _(node: From, ctx: AnnotateContext) -> SQLNode:
    source = node.source
    if isinstance(source, SQLTable):
        # refers to a regular table object, cool
        return FromTable(table=source)
    elif isinstance(source, Symbol):
        # if it is one of the CTE nodes, use that
        over = ctx.cte_nodes.get(source, None)
        if over is not None:
            return FromReference(name=source, over=over)
        else:
            # look into the full table catalog for a table with this name
            table = ctx.catalog.get(source)
            if table is not None:
                return FromTable(table=table)
            else:
                # couldn't resolve the symbol to a known node, sad
                raise ErrRefUndefinedTable(name=source, path=ctx.get_path())

    elif isinstance(source, ValuesTable):
        # A `VALUES` query constructed in the query
        return FromValues(source)
    else:
        return FromNothing()


@annotate_node_scalar.register
def _(node: Fun, ctx: AnnotateContext) -> SQLNode:
    """Annotate all the arguments to the function node"""
    args_p = annotate_scalar(node.args, ctx)
    return Fun(node._name, *args_p)


@annotate_node_scalar.register
def _(node: Get, ctx: AnnotateContext) -> SQLNode:
    return rebind(node.over, Get(name=node._name), ctx)


@annotate_node.register
def _(node: Group, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate(node.over, ctx)
    by_p = annotate_scalar(node.by, ctx)
    return Group(*by_p, over=over_p)


@annotate_node.register
def _(node: Iterate, ctx: AnnotateContext) -> SQLNode:
    """
    The query,
    * `q = base_query >> Iterate (From(alias) >> loop_query >> As(alias))`
    gets translated into:
    * `inter = base_query >> Knot (inter >> loop_query)`
    * `q = inter >> IntIterate()`
    """
    over_p = annotate(node.over, ctx)
    knot = Knot(iterator=node.iterator, box=Box(), over=over_p)
    ctx.mark_origin(knot)

    knot.box.over = knot
    ctx.boxes.append(knot.box)
    ctx.mark_origin(knot.box)

    with ctx.extend_cte_nodes({knot.name: knot.box}):
        start_at = len(ctx.boxes)
        iterator_p = annotate(node.iterator, ctx)
    knot.iterator = iterator_p
    knot.iterator_boxes = ctx.boxes[start_at:]
    return IntIterate(name=label(node.over), iterator_name=knot.name, over=knot.box)


@annotate_node.register
def _(node: Join, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate(node.over, ctx)
    joinee_p = annotate(node.joinee, ctx)
    on_p = annotate_scalar(node.on, ctx)
    return IntJoin(
        joinee=joinee_p,
        on=on_p,
        left=node.left,
        right=node.right,
        skip=node.skip,
        over=over_p,
    )


@annotate_node.register
def _(node: Limit, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate(node.over, ctx)
    return Limit(limit=node.limit, offset=node.offset, over=over_p)


@annotate_node_scalar.register
def _(node: Lit, ctx: AnnotateContext) -> SQLNode:
    return node


@annotate_node.register
def _(node: Order, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate(node.over, ctx)
    by_p = annotate_scalar(node.by, ctx)
    return Order(*by_p, over=over_p)


@annotate_node.register
def _(node: Partition, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate(node.over, ctx)
    by_p = annotate_scalar(node.by, ctx)
    order_by_p = annotate_scalar(node.order_by, ctx)
    return Partition(*by_p, order_by=order_by_p, over=over_p, frame=node.frame)


@annotate_node.register
def _(node: Select, ctx: AnnotateContext) -> SQLNode:
    """
    NOTE: the `label_map` value doesn't need to be updated since it just maps
    the column names to an index in the `args`.
    """
    over_p = annotate(node.over, ctx)
    args_p = annotate_scalar(node.args, ctx)
    return Select(*args_p, over=over_p)


@annotate_node_scalar.register
def _(node: Sort, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate_scalar(node.over, ctx)
    return Sort(value=node.value, nulls=node.nulls, over=over_p)


@annotate_node_scalar.register
def _(node: Var, ctx: AnnotateContext) -> SQLNode:
    return node


@annotate_node.register
def _(node: Where, ctx: AnnotateContext) -> SQLNode:
    over_p = annotate(node.over, ctx)
    condition_p = annotate_scalar(node.condition, ctx)
    return Where(condition=condition_p, over=over_p)


@annotate_node.register
def _(node: With, ctx: AnnotateContext) -> SQLNode:
    args_p = annotate(node.args, ctx)
    more_ctes = {name: args_p[val] for name, val in node.label_map.items()}
    with ctx.extend_cte_nodes(more_ctes):
        over_p = annotate(node.over, ctx)

    return With(*args_p, materialized=node.materialized, over=over_p)



"""This module hosts commonly used utility routines with FunSQL. 
"""

import sqlite3 as sql
from typing import Any

__all__ = ["TableRelations"]


class TableRelations:
    """Keep a cache of all table relationships. Has to be filled in manually."""

    conn: sql.Connection

    def __init__(self):
        self.conn = sql.connect(":memory")
        with self.conn:
            self.conn.execute("DROP TABLE IF EXISTS relations;")
            self.conn.execute(
                "CREATE TABLE relations (TabA TEXT, ColA TEXT, TabB TEXT, ColB TEXT);"
            )

    def add(self, tabA: str, colA: str, tabB: str, colB: str) -> None:
        with self.conn:
            pair = [(tabA, colA, tabB, colB), (tabB, colB, tabA, colA)]
            for values in pair:
                self.conn.execute("INSERT INTO relations VALUES (?, ?, ?, ?);", values)

    def get(self, tabA: str, tabB: str) -> list[Any]:
        return self.conn.execute(
            "SELECT ColA, ColB FROM relations WHERE TabA = ? AND TabB = ?;",
            (tabA, tabB),
        ).fetchall()



from enum import Enum
from typing import Optional, Union

from .nodes import SQLNode
from .clauses import SQLClause
from .sqlcontext import *
from .compiler.annotate import annotate, AnnotateContext
from .compiler.resolve import *
from .compiler.link import *
from .compiler.translate import *
from .compiler.serialize import serialize, SerializationContext, SQLString

# -----------------------------------------------------------
# utility functions to generate SQL context for specific dialects
# -----------------------------------------------------------


def dialect_default() -> SQLDialect:
    return dialect_postgres()


def dialect_postgres() -> SQLDialect:
    return SQLDialect(
        name="postgresql",
        var_style=VarStyle.NUMBERED,
        var_prefix="$",
    )


# -----------------------------------------------------------
# render routines
# -----------------------------------------------------------


class RenderDepth(Enum):
    """
    FunSQL compiles the specified query node structure to a SQL string in multiple passes.
    RenderDepth specifies the number of passes executed, to help with debugging.

    Implementation of the Ordered enumeration is copied from the [reference](https://docs.python.org/3/library/enum.html#orderedenum)
    """

    ANNOTATE = 1
    RESOLVE = 2
    LINK = 3
    TRANSLATE = 4
    SERIALIZE = 5

    def __lt__(self, other):
        if isinstance(other, RenderDepth):
            return self.value < other.value
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, RenderDepth):
            return self.value <= other.value
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, RenderDepth):
            return self.value > other.value
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, RenderDepth):
            return self.value >= other.value
        return NotImplemented


def render(
    node: SQLNode,
    depth: RenderDepth = RenderDepth.SERIALIZE,
    catalog: Optional[SQLCatalog] = None,
) -> Union[SQLNode, SQLClause, SQLString]:
    """
    Render the SQL node expression to a SQLString object.

    Args:
        node: the SQL node to render.
        depth: num of compiler passes to run
        catalog: the SQL catalog to use for resolving SQL table references, and query dialect

    Returns:
        SQLString If all the compiler passes are made, SQLClause object if translate pass has been made, 
        else a SQLNode object. 

    NOTE: calling the `render` method might mutate the SQLNode objects. Work with
    a fresh object on each call.
    """
    assert isinstance(node, SQLNode)
    if catalog is None:
        catalog = SQLCatalog(dialect=dialect_default())

    ann_ctx = AnnotateContext(catalog=catalog)
    node_annotated = annotate(node, ann_ctx)
    if not depth > RenderDepth.ANNOTATE:
        return node_annotated

    resolve_toplevel(ann_ctx)
    if not depth > RenderDepth.RESOLVE:
        return node_annotated

    link_toplevel(ann_ctx)
    if not depth > RenderDepth.LINK:
        return node_annotated

    translate_ctx = TranslateContext(ann_ctx)
    output_clause = translate_toplevel(node_annotated, translate_ctx)
    if not depth > RenderDepth.TRANSLATE:
        return output_clause

    serialize_ctx = SerializationContext(dialect=catalog.dialect)
    serialize(output_clause, serialize_ctx)
    return serialize_ctx.render()


def render_clause(clause: SQLClause, dialect: SQLDialect) -> SQLString:
    """Render the SQL clause to a SQLString object."""
    assert isinstance(clause, SQLClause)
    assert isinstance(dialect, SQLDialect)
    serialize_ctx = SerializationContext(dialect=dialect)
    serialize(clause, serialize_ctx)
    return serialize_ctx.render()
